/* This software is contributed or developed by KYOCERA Corporation. */
/* (C) 2012 KYOCERA Corporation                                      */

#include <linux/module.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/miscdevice.h>
#include <linux/workqueue.h>
#include <linux/uaccess.h>

#include <mach/msm_smd.h>

#include <mach/smem_log.h>

#include "smd_private.h"
#include "smem_fota.h"

#define SMEM_FOTA_DEBUG(...) printk(__VA_ARGS__)

struct fota_device_t {
	struct miscdevice misc;
};

spinlock_t smem_fota_lock;

static int fota_ioctl(struct inode *inode, struct file *file,
				unsigned int cmd, unsigned long arg)
{
	struct smem_fota_ioctl_info info;
	unsigned char *ptr = NULL;
	unsigned long flags;
	unsigned long result = -EFAULT;
	switch(cmd)
	{
		case IOCTL_SMEM_FOTA_READ_CMD:
			result = copy_from_user(&info,(struct smem_fota_ioctl_info __user *)arg,sizeof(info));
			SMEM_FOTA_DEBUG("@@@DEBUG IOCTL_SMEM_FOTA_READ_CMD offset[%d] size[%d]\n",info.offset,info.size);
			ptr = (unsigned char *)smem_alloc(SMEM_FOTA_INFO, SMEM_FOTA_INFO_SIZE);
			if(ptr==0)
			{
				info.size = 0;
				SMEM_FOTA_DEBUG("@@@DEBUG IOCTL_SMEM_FOTA_READ_CMD ERROR!!!\n");
			}
			else
			{
				if(info.offset+info.size > SMEM_FOTA_INFO_SIZE)
				{
					info.size = SMEM_FOTA_INFO_SIZE - info.offset;
				}
				spin_lock_irqsave(&smem_fota_lock, flags);
				memcpy(info.data, (ptr+info.offset), info.size);
				spin_unlock_irqrestore(&smem_fota_lock, flags);
			}
			SMEM_FOTA_DEBUG("@@@DEBUG IOCTL_SMEM_FOTA_READ_CMD [%d][%d][%d]\n",info.data[0],info.data[1],info.data[2]);
			result = copy_to_user((struct smem_fota_ioctl_info __user *)arg,&info,sizeof(info));
			break;

		case IOCTL_SMEM_FOTA_WRITE_CMD:
			result = copy_from_user(&info,(struct smem_fota_ioctl_info __user *)arg,sizeof(info));
			SMEM_FOTA_DEBUG("@@@DEBUG IOCTL_SMEM_FOTA_WRITE_CMD offset[%d] size[%d]\n",info.offset,info.size);
			ptr = (unsigned char *)smem_alloc(SMEM_FOTA_INFO, SMEM_FOTA_INFO_SIZE);
			if(ptr==0)
			{
				info.size = 0;
				SMEM_FOTA_DEBUG("@@@DEBUG IOCTL_SMEM_FOTA_WRITE_CMD ERROR!!!\n");
			}
			else
			{
				if(info.offset+info.size > SMEM_FOTA_INFO_SIZE)
				{
					info.size = SMEM_FOTA_INFO_SIZE - info.offset;
				}
				spin_lock_irqsave(&smem_fota_lock, flags);
				memcpy((ptr+info.offset), info.data, info.size);
				spin_unlock_irqrestore(&smem_fota_lock, flags);
			}
			result = copy_to_user((struct smem_fota_ioctl_info __user *)arg,&info,sizeof(info));
			break;
		default:
			break;
	}
	return 0;
}

static int fota_open(struct inode *ip, struct file *fp)
{
	return nonseekable_open(ip, fp);
}

static int fota_release(struct inode *ip, struct file *fp)
{
	return 0;
}

static const struct file_operations fota_fops = {
	.owner = THIS_MODULE,
	.open = fota_open,
	.ioctl = fota_ioctl,
	.release = fota_release,
};

static struct fota_device_t fota_device = {
	.misc = {
		.minor = MISC_DYNAMIC_MINOR,
		.name = "fota_sm",
		.fops = &fota_fops,
	}
};

static void __exit fota_exit(void)
{
	misc_deregister(&fota_device.misc);
}

static int __init fota_init(void)
{
	int ret;
	ret = misc_register(&fota_device.misc);
	return ret;
}

module_init(fota_init);
module_exit(fota_exit);

MODULE_DESCRIPTION("MSM Shared Memory FOTA Driver");
MODULE_LICENSE("GPL v2");
