/*
 * This software is contributed or developed by KYOCERA Corporation.
 * (C) 2011 KYOCERA Corporation
*/


#include <linux/delay.h>
#include <linux/types.h>
#include <linux/i2c.h>
#include <linux/uaccess.h>
#include <linux/miscdevice.h>
#include <media/msm_camera.h>
#include <mach/gpio.h>
#include <mach/kcj_dev_info.h>
#include <linux/slab.h>
#include "nt99250.h"
#include "../../../arch/arm/mach-msm/proc_comm.h"
#include <linux/sched.h>
#include <mach/hs_io_ctl_a.h>
#include <linux/kthread.h>

#define SEQ_LOG_SWITCH
#undef SEQ_LOG_SWITCH
#ifdef SEQ_LOG_SWITCH
#define SEQ_LOG(fmt, args...) printk(KERN_INFO fmt, ##args)
#else
#define SEQ_LOG(fmt, args...) do{ } while(0)
#endif

#define API_LOG_SWITCH
#ifdef API_LOG_SWITCH
#define API_LOG(fmt, args...) printk(KERN_INFO fmt, ##args)
#else
#define API_LOG(fmt, args...) do{ } while(0)
#endif

#define DBG_LOG_SWITCH
#ifdef DBG_LOG_SWITCH
#define DBG_LOG(fmt, args...) printk(KERN_NOTICE fmt, ##args)
#else
#define DBG_LOG(fmt, args...) do{ } while(0)
#endif

#define ERR_LOG_SWITCH
#ifdef ERR_LOG_SWITCH
#define ERR_LOG(fmt, args...) printk(KERN_ERR fmt, ##args)
#else
#define ERR_LOG(fmt, args...) do{ } while(0)
#endif

#define DBG_TIME_LOG_SWITCH
#undef DBG_TIME_LOG_SWITCH
#ifdef DBG_TIME_LOG_SWITCH
#define DBG_TIME_LOG(fmt, args...) printk(KERN_INFO fmt, ##args)
#else
#define DBG_TIME_LOG(fmt, args...) do{ } while(0)
#endif

#define  REG_NT99250_MODEL_ID 0x3000
#define  NT99250_MODEL_ID     0x2500

#define  REG_NT99250_SENSOR_RESET     0x3021


#define NT99250_REG_EFFECTS_0             0x32F1
#define NT99250_REG_EFFECTS_1             0x32F6

#define NT99250_EFFECTS_NO_EFFECT       0x0000
#define NT99250_EFFECTS_MONOCHROME      0x0001
#define NT99250_EFFECTS_NEGATIVE_MONO   0x0002
#define NT99250_EFFECTS_NEGATIVE_COLOR  0x0003
#define NT99250_EFFECTS_SEPIA           0x0004


#define NT99250_REG_FLICKERQUANT        0x04BA

#define NT99250_FLICKE_AUTO             0x0000
#define NT99250_FLICKE_50HZ             0x0001
#define NT99250_FLICKE_60HZ             0x0002

#define NT99250_REG_RGAIN               0x20A4
#define NT99250_REG_BGAIN               0x20A6


#define NT99250_REG_OTP_USE_EFUSE       0x0688



#define NT99250_DEFAULT_CLOCK_RATE  24000000
#define I2C_DEVICE_ID_I2C_CAMERA  2

#ifdef CONFIG_MSM_CAMERA_FLASH
#if defined(CONFIG_TARGET_PROTOTYPE_WS0) | defined(CONFIG_TARGET_PROTOTYPE_WS1)
#else
#define I2C_DEVICE_AS3685C
#define I2C_DEVICE_ID_I2C_FLASH   4
#define I2C_DEVICE_NAME_FLASH   "as3685c"
#endif
#endif

#define AS3685C_SLAVE_ADDRESS   (0x82 >> 1)

#define NT99250_CHIP_ID                0x25
#define NT99250_CHIP_REV               0x00

#define FPS_30 30
#define FPS_24 24
#define FPS_15 15
#define FPS_10 10
#define FPS_7P5 7
#define FPS_5   5

#define NT99250_SCENE_COMMON_BRIGHTNESS 4
#define NT99250_SCENE_COMMON_EV 0


#define CAM_DRV_CURRENT_THREAD_WAIT 0
#define CAM_DRV_THREAD_WAKE_UP      1

#define CAMERA_INT_VSYNC_NONE                 0x0000
#define CAMERA_INT_VSYNC_HIGH                 0x0001
#define CAMERA_INT_VSYNC_LOW                  0x0002
#define CAMERA_INT_VSYNC_ALWAYS               0x0004

#define CAMERA_PROC_NONE                      0x0000

#define CAMERA_PROC_EXPOSURE                  0x0001
#define CAMERA_PROC_AF_MODE                   0x0002
#define CAMERA_PROC_STABLE_SEARCH             0x0004

#define CAMERA_PROC_REGSTER_WRITE             0x0008
#define CAMERA_PROC_WAIT_V_FRAME              0x0010

#define CAMERA_PROC_FLASH_STILL               0x0020
#define CAMERA_PROC_FLASH_ON                  0x0040
#define CAMERA_PROC_FLASH_OFF                 0x0080

#define CAMERA_PROC_CONVERGENCE_WB            0x0100
#define CAMERA_PROC_ALWAYS                    0x1000
#define CAMERA_PROC_DUMMY                     0x8000

#define CAM_MSG_QUE_MAX        16

#define NT99250_AWB_GAIN_WAIT_TIME            1000
#define NT99250_AWB_GAIN_READ_CNT             5


#define CAM_DVDD_CNT_ENABLE
#undef  CAM_DVDD_CNT_ENABLE
#define VREG_PULL_DOWN_SETTING
#undef  VREG_PULL_DOWN_SETTING

#define STILL_CONTINUANCE_FLG
#undef  STILL_CONTINUANCE_FLG

typedef enum
{
  CAMERA_FLASH_CTRL_OFF,
  CAMERA_FLASH_CTRL_MAIN,
  CAMERA_FLASH_CTRL_PRE,
  CAMERA_FLASH_CTRL_TORCH,
} camera_flash_ctrl_type;

camera_flash_ctrl_type  flash_ctrl_mode;

typedef enum
{
  CAMERA_FLASH_CHK_FALSE,
  CAMERA_FLASH_CHK_TRUE,
  CAMERA_FLASH_CHK_NONE,
} camera_flash_check_type;

typedef enum
{
  CAMERA_IDLE_FOCUS,
  CAMERA_AF_FOCUS,
  CAMERA_MOVE_FOCUS,
} camera_focus_state_type;

typedef enum
{
  ADJUST_WB_STATE_NONE,
  ADJUST_WB_STATE_START,
  ADJUST_WB_STATE_CALC
} camera_wb_adjust_state_type;

typedef enum
{
  CAMERA_TEST_RESULT_OK,
  CAMERA_TEST_RESULT_NG
} camera_check_result_type;

typedef enum
{
  CAMERA_TYPE_INIT = 0,
  CAMERA_TYPE_FRONT,
  CAMERA_TYPE_REAR,
} camera_ctrl_type;

typedef enum
{
  CAMERA_MSG_SYNC_MODE = 0,
  CAMERA_MSG_ASYNC_MODE,
} camera_msg_sync_mode;

typedef enum
{
  CAMERA_MSG_CTRL_INIT = 0,
  CAMERA_MSG_CTRL_SUCCESS,
  CAMERA_MSG_CTRL_FAIL,
  CAMERA_MSG_CTRL_NO_MSG,
  CAMERA_MSG_CTRL_OTHER_ERR,
} camera_msg_err_code;

typedef enum
{
  CAMERA_CTRL_INIT = 0,
  CAMERA_CTRL_SENSOR_MODE,
  CAMERA_CTRL_AF_INIT,
  CAMERA_CTRL_AF_CMD,
  CAMERA_CTRL_WB,
  CAMERA_CTRL_EFFECT,
  CAMERA_CTRL_AEC,
  CAMERA_CTRL_SCENE,
  CAMERA_CTRL_ISO,
  CAMERA_CTRL_ANTIBANDING,
  CAMERA_CTRL_CONTRAST,
  CAMERA_CTRL_BRIGHTNESS,
  CAMERA_CTRL_SHARPNESS,
  CAMERA_CTRL_LENS_SHADING,
  CAMERA_CTRL_FLASH,
  CAMERA_CTRL_MAX,
} camera_cmd_type;

typedef enum
{
  DBG_TIME_INIT = 0,
  DBG_TIME_START,
  DBG_TIME_END,
  DBG_TIME_MAX,
} dbg_time_cmd;


struct cam_msgbuf{
   int  msqid;
   int  msg_type;
   void *msg_data;
   int  msg_size;
   int  msg_flg;
};

struct cam_msgque{
    int snd_idx;
    int rcv_idx;
    int list_num;
    struct cam_msgbuf pst_msg[CAM_MSG_QUE_MAX];
};

static struct cam_msgque *pst_cam_msgque;

struct nt99250_ctrl {
    const struct msm_camera_sensor_info *sensordata;

    int wait_flag;
    int wait_thread_flag;

    struct nt99250_reg_t *nt99250_regs;
    uint16_t  hw_ver[2];
    int sensormode;

    uint8_t  flash_mode;
    uint8_t  af_mode;
    uint8_t  brightness_mode;
    uint8_t  contrast_mode;
    uint8_t  sharpness_mode;
    uint8_t  scene_mode;
    uint8_t  effect_mode;
    uint8_t  wb_mode;
    uint8_t  iso_mode;
    uint8_t  aec_mode;
    uint8_t  antibanding_mode;
    uint8_t  lens_shading_mode;
    int flash_on_flg;
    uint32_t exposure_time;
    uint16_t brightness;
    uint16_t analog_gain;
    unsigned long nt99250_otp0_H;
    unsigned long nt99250_otp0_M;
    unsigned long nt99250_otp0_L;
    unsigned long nt99250_otp1_H;
    unsigned long nt99250_otp1_M;
    unsigned long nt99250_otp1_L;

    uint16_t v_entry_proc;
    uint16_t v_int_kind;
    camera_wb_adjust_state_type camera_wb_adjust_state;
    uint8_t  wb_adjust_cnt;
    unsigned short rgain_data[NT99250_AWB_GAIN_READ_CNT];
    unsigned short bgain_data[NT99250_AWB_GAIN_READ_CNT];

    enum msm_s_resolution prev_res;
    enum msm_s_resolution pict_res;
    enum msm_s_resolution curr_res;
    enum msm_s_test_mode  set_test;
    struct task_struct *camdev_th;
};

struct nt99250_work {
    struct work_struct work;
};

#ifdef I2C_DEVICE_AS3685C
struct flash_work {
    struct work_struct work;
};
#endif

#ifdef FEATURE_KYOCERA_MCAM_BB_ADJ
typedef enum
{
  NT99250_KY_ADJUST_INIT,
  NT99250_KY_ADJUST_FLASH,
  NT99250_KY_ADJUST_SCENE_AUTO,
  NT99250_KY_ADJUST_SCENE_PORTRAIT,
  NT99250_KY_ADJUST_SCENE_SCENERY,
  NT99250_KY_ADJUST_SCENE_NIGHT_PORTRAIT,
  NT99250_KY_ADJUST_SCENE_NIGHT_SCENERY,
  NT99250_KY_ADJUST_SCENE_ACTION,
  NT99250_KY_ADJUST_MAX,
}nt99250_ky_adjust_type;

typedef struct{
  unsigned long  partitionCode;
  unsigned long  dat_blk_No;
  unsigned long  regAdr;
  unsigned long  regDat;
}nt99250_Blk_struct;
#endif

#if 0
static camera_scene_info_type scene_info[ CAM_SCENE_MAX ] = 
{
  { 
    CAM_WB_AUTO,
    CAM_AF_MODE_AUTO,
    CAM_AEC_CENTER,
    CAM_ISO_AUTO,
    CAM_FLASH_OFF,
    NT99250_SCENE_COMMON_BRIGHTNESS,
    CAM_ANTIBANDING_60HZ,
    NT99250_SCENE_COMMON_EV,
    FPS_24,
    3,
    3,
    3,
    CAM_EFFECT_NONE,
    TRUE
  },
  { 
    CAM_WB_AUTO,
    CAM_AF_MODE_AUTO,
    CAM_AEC_SPOT,
    CAM_ISO_AUTO,
    CAM_FLASH_OFF,
    NT99250_SCENE_COMMON_BRIGHTNESS,
    CAM_ANTIBANDING_60HZ,
    NT99250_SCENE_COMMON_EV,
    FPS_24,
    2,
    3,
    3,
    CAM_EFFECT_NONE,
    TRUE
  },
  { 
    CAM_WB_AUTO,
    CAM_AF_MODE_AUTO,
    CAM_AEC_AVERAGE,
    CAM_ISO_AUTO,
    CAM_FLASH_OFF,
    NT99250_SCENE_COMMON_BRIGHTNESS,
    CAM_ANTIBANDING_60HZ,
    NT99250_SCENE_COMMON_EV,
    FPS_24,
    4,
    4,
    4,
    CAM_EFFECT_NONE,
    TRUE
  },
  { 
    CAM_WB_AUTO,
    CAM_AF_MODE_AUTO,
    CAM_AEC_SPOT,
    CAM_ISO_AUTO,
    CAM_FLASH_OFF,
    NT99250_SCENE_COMMON_BRIGHTNESS,
    CAM_ANTIBANDING_60HZ,
    NT99250_SCENE_COMMON_EV,
    FPS_24,
    2,
    3,
    3,
    CAM_EFFECT_NONE,
    TRUE
  },
  { 
    CAM_WB_AUTO,
    CAM_AF_MODE_NORMAL,
    CAM_AEC_SPOT,
    CAM_ISO_AUTO,
    CAM_FLASH_OFF,
    NT99250_SCENE_COMMON_BRIGHTNESS,
    CAM_ANTIBANDING_60HZ,
    NT99250_SCENE_COMMON_EV,
    FPS_24,
    3,
    3,
    3,
    CAM_EFFECT_NONE,
    TRUE
  },
  { 
    CAM_WB_AUTO,
    CAM_AF_MODE_NORMAL,
    CAM_AEC_SPOT,
    CAM_ISO_400,
    CAM_FLASH_OFF,
    NT99250_SCENE_COMMON_BRIGHTNESS,
    CAM_ANTIBANDING_60HZ,
    NT99250_SCENE_COMMON_EV,
    FPS_24,
    3,
    3,
    3,
    CAM_EFFECT_NONE,
    TRUE
  },
};
#endif 
static struct nt99250_work *nt99250_sensorw;
static struct i2c_client    *nt99250_client;
static struct nt99250_ctrl *nt99250_ctrl;

#ifdef I2C_DEVICE_AS3685C
static struct flash_work *flash_sensorw;
static struct i2c_client *flash_client;
#endif

static struct nt99250_i2c_reg_conf
nt99250_set_effect[] = 
{
{ 0x0028, 0x7000, WORD_LEN, 0 },
{ 0x002A, 0x021E, WORD_LEN, 0 },
{ 0x0F12, 0x0003, WORD_LEN, 0 },
};

#if 0
static struct nt99250_i2c_reg_conf
nt99250_set_iso[] = 
{
{ 0x0028, 0x7000, WORD_LEN, 0 },
{ 0x002A, 0x04B4, WORD_LEN, 0 },
{ 0x0F12, 0x0000, WORD_LEN, 0 },
{ 0x0F12, 0x0000, WORD_LEN, 0 },
{ 0x0F12, 0x0001, WORD_LEN, 0 },
};
#endif
static DECLARE_WAIT_QUEUE_HEAD(nt99250_wait_queue);
DEFINE_MUTEX(nt99250_mutex);
DEFINE_MUTEX(nt99250_msg_mutex);
#ifdef I2C_DEVICE_AS3685C
static DECLARE_WAIT_QUEUE_HEAD(flash_ctrl_wait_queue);
#endif
static DECLARE_WAIT_QUEUE_HEAD(cam_dev_wait_queue);


#if 0
#ifdef I2C_DEVICE_AS3685C
static int L_flash_ctrl_i2c_init_client( struct i2c_client *client );
static int L_flash_ctrl_i2c_probe( struct i2c_client *client, const struct i2c_device_id *id );
static int __exit L_flash_ctrl_i2c_remove( struct i2c_client *client );

static int32_t L_flash_ctrl_i2c_txdata( unsigned short saddr, unsigned char *txdata, int length );
#endif
static int32_t L_flash_ctrl_i2c_write( unsigned short saddr, unsigned char data1, unsigned char data2, unsigned char data3 );
#endif

static int L_nt99250_i2c_init_client( struct i2c_client *client );
static int L_nt99250_i2c_probe( struct i2c_client *client, const struct i2c_device_id *id );
static int __exit L_nt99250_i2c_remove( struct i2c_client *client );

static int L_nt99250_i2c_rxdata( unsigned short saddr, unsigned char *rxdata, int length );
static int32_t L_nt99250_i2c_txdata( unsigned short saddr,unsigned char *txdata,int length );
static int32_t L_nt99250_i2c_write( unsigned short saddr, unsigned short waddr, unsigned long wdata, enum nt99250_width_t width);
static int32_t L_nt99250_i2c_write_table( struct nt99250_i2c_reg_conf const *reg_cfg_tbl, int num );
static int32_t L_nt99250_i2c_read( unsigned short saddr, unsigned short raddr, unsigned short *rdata, enum nt99250_width_t width );

static int L_nt99250_reset(const struct msm_camera_sensor_info *dev);
static int L_nt99250_probe_init_sensor( const struct msm_camera_sensor_info *data );
static int L_nt99250_set_reg_table( void );
static int L_nt99250_sensor_init( const struct msm_camera_sensor_info *data );
static int L_nt99250_sensor_release( void );
static int L_nt99250_sensor_config( void __user *argp );
static int L_nt99250_sensor_probe( const struct msm_camera_sensor_info *info, struct msm_sensor_ctrl *sensor );
static int L_nt99250_probe( struct platform_device *pdev );

static int32_t L_nt99250_setting( enum msm_s_reg_update rupdate,enum msm_s_setting rsetting );
static void L_nt99250_pwd_config( unsigned on_off );
static int32_t L_nt99250_power_down( void );
static int32_t L_nt99250_video_config( int mode, int res );
static int32_t L_nt99250_snapshot_config( int mode );
static int32_t L_nt99250_raw_snapshot_config( int mode );
static int32_t L_nt99250_prepare_snapshot( void );

static int L_nt99250_set_effect( camera_effect_type effect );
static int32_t L_nt99250_set_sensor_mode( int mode, int res );

static void L_nt99250_entry_vsync_proc( uint8_t cmd, uint16_t entry );
static void L_nt99250_delete_vsync_proc( uint16_t entry );
static void L_nt99250_proc_vsync( unsigned long data );

static irqreturn_t L_nt99250_vsync_interrupt( int irq, void *handle );

static int L_nt99250_wake_up( void );
static int L_nt99250_wait_event( void );


static int L_nt99250_thread_init( void );
static void L_nt99250_thread_release( void );
static int L_nt99250_msg_proc( struct cam_msgbuf *msgp );
static int L_nt99250_msg_get( void );
static int L_nt99250_msg_send
(
  struct cam_msgbuf *msgp,
  int msgflg
);
static camera_msg_err_code L_nt99250_msg_rcv
(
  struct cam_msgbuf *msgp,
  int msgtyp,
  int msgflg
);
static int L_nt99250_register_accsess
(
  camera_cmd_type cmd,
  void* data,
  uint16_t size
);
static int L_nt99250_thread_wait_event(void);

static int L_nt99250_thread_should_stop(void);
static int L_nt99250_thread_run(void);
static int L_nt99250_thread_stop(void);
static int L_nt99250_thread_wake_up( void );
static int L_nt99250_thread(void *data);

static int L_nt99250_read_hw_ver( void );


#ifdef FEATURE_KYOCERA_MCAM_BB_ADJ
#endif

static int L_nt99250_get_devid( camera_devid_info* devid_info );
static int L_nt99250_devid_check( uint8_t* ver_check_result );

#ifdef DBG_TIME_LOG_SWITCH
static void L_nt99250_dbg_proc_time(dbg_time_cmd cmd_type);
#endif

#if 0
#ifdef I2C_DEVICE_AS3685C
static const struct i2c_device_id flash_ctrl_i2c_id[] = {
    { I2C_DEVICE_NAME_FLASH, I2C_DEVICE_ID_I2C_FLASH},
    { },
};

static int L_flash_ctrl_i2c_init_client
(
  struct i2c_client *client
)
{
    API_LOG("%s E \n", __func__);

    init_waitqueue_head(&flash_ctrl_wait_queue);

    API_LOG("%s X \n", __func__);

    return 0;
}

static int L_flash_ctrl_i2c_probe
(
  struct i2c_client *client,
  const struct i2c_device_id *id
)
{
    int rc = 0;

    API_LOG("%s E \n", __func__);

    if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C))
    {
        kfree(flash_sensorw);
        flash_sensorw = NULL;
        return rc;
    }

    flash_sensorw = kzalloc(sizeof(struct flash_work), GFP_KERNEL);
    if (!flash_sensorw)
    {
        rc = -ENOMEM;
        kfree(flash_sensorw);
        flash_sensorw = NULL;
        return rc;
    }

    i2c_set_clientdata(client, flash_sensorw);
    L_flash_ctrl_i2c_init_client(client);
    flash_client = client;

    mdelay(50);

    API_LOG("%s X \n", __func__);

    return rc;
}

static int __exit L_flash_ctrl_i2c_remove
(
  struct i2c_client *client
)
{
    struct flash_work_t *sensorw = i2c_get_clientdata(client);

    API_LOG("%s E \n", __func__);

    free_irq(client->irq, sensorw);
    flash_client = NULL;
    flash_sensorw = NULL;
    kfree(sensorw);

    API_LOG("%s X \n", __func__);

    return 0;
}

static struct i2c_driver flash_ctrl_i2c_driver = {
    .id_table = flash_ctrl_i2c_id,
    .probe  = L_flash_ctrl_i2c_probe,
    .remove = __exit_p(L_flash_ctrl_i2c_remove),
    .driver = {
        .name = I2C_DEVICE_NAME_FLASH,
    },
};

static int32_t L_flash_ctrl_i2c_txdata
(
  unsigned short saddr,
  unsigned char *txdata, 
  int length
)
{
    struct i2c_msg msg[] = {
        {
          saddr, 0, length, txdata,
        },
    };
    int ret;
#ifdef CAMERA_I2C_IRQ_DISABLE
    unsigned long flags;
#endif

    API_LOG("%s E \n", __func__);

#ifdef CAMERA_I2C_IRQ_DISABLE
    local_irq_save(flags);

    local_irq_disable();
    ret = i2c_transfer(flash_client->adapter, msg, 1);
    if(ret < 0)
    {
        local_irq_enable();
        local_irq_restore(flags);
        ERR_LOG("%s failed\n",__func__);
        return -EIO;
    }

    local_irq_enable();
    local_irq_restore(flags);
#else
    ret = i2c_transfer(flash_client->adapter, msg, 1);
    if (ret < 0)
    {
        ERR_LOG("%s failed\n",__func__);
        return -EIO;
    }
#endif

    API_LOG("%s X \n", __func__);

    return 0;
}
#endif

static int32_t L_flash_ctrl_i2c_write
(
  unsigned short saddr,
  unsigned char data1,
  unsigned char data2,
  unsigned char data3
)
{
    int32_t rc = TRUE;
    unsigned char buf[3];

    API_LOG("%s E \n", __func__);

    memset(buf, 0, sizeof(buf));
  
    buf[0] = data1;
    buf[1] = data2;
    buf[2] = data3;

    DBG_LOG("saddr   = [%04x] \n",saddr);
    DBG_LOG(" buf = [%02x",buf[0]);
    DBG_LOG(" %02x",buf[1]);
    DBG_LOG(" %02x]\n",buf[2]);

#ifdef I2C_DEVICE_AS3685C
    rc = L_flash_ctrl_i2c_txdata(saddr, buf, 3);
#else
    rc = L_nt99250_i2c_txdata(saddr, buf, 3);
#endif
    if (rc < 0)
    {
        ERR_LOG("Err: data1:0x%2x data2:0x%2x data3:0x%2x \n", data1, data2, data3 );
    }
    else
    {
        SEQ_LOG("Success data1:0x%2x data2:0x%2x data3:0x%2x \n", data1, data2, data3 );
    }

    API_LOG("%s X \n", __func__);

    return rc;
}

static int L_flash_ctrl_threshold_check
(
  camera_flash_check_type* flash_onoff_flg
)
{
    int rc = TRUE;
    uint16_t Again = 0;
    camera_flash_check_type flash_flg;
    uint8_t  current_flash_mode;

    API_LOG("%s E \n", __func__);

    if( flash_onoff_flg == NULL )
    {
        ERR_LOG("%s Param NULL Err\n",__func__);
        return -EIO;
    }

    SEQ_LOG("scene_mode (%d)\n",nt99250_ctrl->scene_mode);

    if(nt99250_ctrl->scene_mode != CAM_SCENE_AUTO)
    {
        current_flash_mode = scene_info[ nt99250_ctrl->scene_mode ].flash;
    }
    else
    {
        current_flash_mode = nt99250_ctrl->flash_mode;
    }

    SEQ_LOG("[%s] flash_mode : [%d]\n",__func__ ,current_flash_mode);

    if( current_flash_mode == CAM_FLASH_AUTO )
    {

        rc = L_nt99250_i2c_read_reg_data(NT99250_REG_RAM_BASE_ADDR,CAMSENSOR_REG_ANALOG_GAIN,&Again);
        if( rc < 0 )
        {
            ERR_LOG( "L_flash_ctrl_threshold_check Err\n");
            return rc;
        }

        if( ((unsigned short)Again) & 0x800 )
        {
            flash_flg = CAMERA_FLASH_CHK_TRUE;
            SEQ_LOG("FLASH TRUE\n");
        }
        else
        {
            flash_flg = CAMERA_FLASH_CHK_FALSE;
            SEQ_LOG("FLASH FALSE\n");
        }
    }
    else if(current_flash_mode == CAM_FLASH_ON)
    {
        flash_flg = CAMERA_FLASH_CHK_TRUE;
    }
    else
    {
        flash_flg = CAMERA_FLASH_CHK_FALSE;
    }

    *flash_onoff_flg = flash_flg;

    SEQ_LOG("[%s] flash_check_result : [%d]\n",__func__ ,*flash_onoff_flg);
    API_LOG("%s X \n", __func__);

    return rc;
}

static int L_flash_ctrl_video_light_set
(
  int mode
)
{
    int rc = TRUE;
    uint8_t data1;
    uint8_t data2;
    uint8_t data3;
    struct nt99250_i2c_reg_conf const* set_reg;

    API_LOG("%s E \n", __func__);

    set_reg = &nt99250_regs.set_flash_mode[0];

    data1 = set_reg->wdata;

    set_reg = &nt99250_regs.set_flash_mode[3];

    data2 = set_reg->wdata;

    set_reg = &nt99250_regs.set_flash_mode[2];

    SEQ_LOG("%s mode:[%d]\n",__func__,mode);

    if(mode == CAM_FLASH_TORCH)
    {
        data3 = set_reg->wdata;
        data3 = (data3 | AS3685C_DATA3_TORCH);
    }
    else
    {
        data3 = set_reg->wdata;
    }

    rc = L_flash_ctrl_i2c_write( AS3685C_SLAVE_ADDRESS, data1, data2, data3 );
    if( rc < 0 )
    {
        ERR_LOG("%s Err : %d \n",__func__, rc);
    }

    API_LOG("%s X \n", __func__);

    return rc;
}

static int L_nt99250_flash_control
(
  uint8_t  mode
)
{
    int rc = TRUE;
    uint8_t data1;
    uint8_t data2;
    uint8_t data3;
    struct nt99250_i2c_reg_conf const* set_reg;

    API_LOG("%s E \n", __func__);

    set_reg = &nt99250_regs.set_flash_mode[0];

    data1 = set_reg->wdata;
    set_reg++;

    data2 = set_reg->wdata;
    set_reg++;

    SEQ_LOG("%s mode:[%d]\n",__func__,mode);

    if(mode == CAMERA_FLASH_CTRL_MAIN)
    {
        data3 = set_reg->wdata;
        data3 = (data3 | AS3685C_DATA3_MAIN);
    }
    else if(mode == CAMERA_FLASH_CTRL_PRE)
    {
        data3 = set_reg->wdata;
        data3 = (data3 | AS3685C_DATA3_TORCH);
    }
    else
    {
        data3 = set_reg->wdata;
    }

    rc = L_flash_ctrl_i2c_write( AS3685C_SLAVE_ADDRESS, data1, data2, data3 );
    if( rc < 0 )
    {
        ERR_LOG("%s Err\n",__func__);
    }
    
    API_LOG("%s X \n", __func__);

    return rc;
}

int nt99250_flash_set_led_mode
(
  int flash_mode
)
{
    int rc = TRUE;

    API_LOG("%s E mode:[%d]", __func__, flash_mode);

    if( nt99250_ctrl->flash_mode == flash_mode )
    {
        SEQ_LOG("Same Flash Mode Value \n");
        return TRUE;
    }

    if( ((nt99250_ctrl->flash_mode == CAM_FLASH_TORCH) && (flash_mode != CAM_FLASH_TORCH))
      ||((nt99250_ctrl->flash_mode != CAM_FLASH_TORCH) && (flash_mode == CAM_FLASH_TORCH))
      )
    {
        rc = L_flash_ctrl_video_light_set(flash_mode);
        SEQ_LOG("video_light rc:[%d] \n",rc);
    }

    nt99250_ctrl->flash_mode = flash_mode;

    API_LOG("%s X \n", __func__);

    return rc;
}
#endif
static int L_nt99250_i2c_rxdata
(
  unsigned short saddr,
  unsigned char *rxdata,
  int length
)
{
    struct i2c_msg msgs[] = {
        {
          saddr, 0, 2, rxdata,
        },
        {
          saddr, I2C_M_RD, 1, rxdata,
        },
    };
    int ret;
#ifdef CAMERA_I2C_IRQ_DISABLE
    unsigned long flags;
#endif

    API_LOG("%s E \n", __func__);

#ifdef CAMERA_I2C_IRQ_DISABLE
    local_irq_save(flags);

    local_irq_disable();
    ret = i2c_transfer(nt99250_client->adapter, msgs, 2);
    if(ret < 0)
    {
        local_irq_enable();
        local_irq_restore(flags);
        ERR_LOG("%s failed!  ret:[%d]\n",__func__ ,ret);
        return -EIO;
    }

    local_irq_enable();
    local_irq_restore(flags);
#else
    API_LOG("%s E adapter name =%s\n", __func__,nt99250_client->adapter->name);

    ret = i2c_transfer(nt99250_client->adapter, msgs, 2);
    if (ret < 0)
    {
        ERR_LOG("%s failed!  ret:[%d]\n",__func__ ,ret);
        return -EIO;
    }
#endif

    API_LOG("%s X \n", __func__);

    return 0;
}

static int32_t L_nt99250_i2c_txdata
(
  unsigned short saddr,
  unsigned char *txdata, 
  int length
)
{
    struct i2c_msg msg[] = {
        {
          saddr, 0, length, txdata,
        },
    };
    int ret;
#ifdef CAMERA_I2C_IRQ_DISABLE
    unsigned long flags;
#endif

    API_LOG("%s E \n", __func__);

#ifdef CAMERA_I2C_IRQ_DISABLE
    local_irq_save(flags);

    local_irq_disable();
    ret = i2c_transfer(nt99250_client->adapter, msg, 1);
    if(ret < 0)
    {
        local_irq_enable();
        local_irq_restore(flags);
        ERR_LOG("L_nt99250_i2c_txdata failed ret:[%d]\n",ret);
        return -EIO;
    }

    local_irq_enable();
    local_irq_restore(flags);
#else
    ret = i2c_transfer(nt99250_client->adapter, msg, 1);
    if (ret < 0)
    {
        ERR_LOG("L_nt99250_i2c_txdata failed! ret:[%d]\n",ret);
        return -EIO;
    }
#endif

    API_LOG("%s X \n", __func__);

    return 0;
}

static int32_t L_nt99250_i2c_write
(
  unsigned short saddr,
  unsigned short waddr,
  unsigned long wdata,
  enum nt99250_width_t  width
)
{
    int32_t rc = 0;
    unsigned char buf[6];
    int length;

    API_LOG("%s E \n", __func__);

    memset(buf, 0, sizeof(buf));

    SEQ_LOG("%s width = [%d] \n",__func__,width);

    switch (width)
    {
        case WORD_LEN:
            buf[0] = (unsigned char)((waddr & 0xFF00)>>8);
            buf[1] = (unsigned char)((waddr & 0x00FF));
            buf[2] = (unsigned char)((wdata & 0x00FF));
            length = 3;
            break;
        case BYTE_LEN:
            buf[0] = (unsigned char)(waddr & 0x000000FF);
            buf[1] = (unsigned char)(wdata & 0x000000FF);
            length = 2;
            break;
        default:
            rc = -EIO;
            ERR_LOG("%s Param Err\n", __func__);
            break;
    }

    if (!(rc < 0))
    {
        rc = L_nt99250_i2c_txdata(saddr, buf, length);
        if (rc < 0)
        {
            ERR_LOG("i2c_write_w failed, addr = 0x%x, val = 0x%x!\n",waddr, (unsigned int)wdata);
        }
    }

    API_LOG("%s X \n", __func__);

    return rc;
}

static int32_t L_nt99250_i2c_write_table
(
  struct nt99250_i2c_reg_conf const *reg_cfg_tbl,
  int num
)
{
    int i;
    int32_t rc = -EIO;

    API_LOG("%s E \n", __func__);
    DBG_LOG("addr   = [%04x] \n",nt99250_client->addr);
    DBG_LOG("write num = [%d] \n",num);

    for (i = 0; i < num; i++)
    {

        DBG_LOG("w_addr = [%04x] \n",reg_cfg_tbl->waddr);
        DBG_LOG("w_data = [%04x] \n",(unsigned int)reg_cfg_tbl->wdata);

        rc = L_nt99250_i2c_write(nt99250_client->addr,
            reg_cfg_tbl->waddr, reg_cfg_tbl->wdata, reg_cfg_tbl->width);
        if (rc < 0)
        {
            ERR_LOG("L_nt99250_i2c_write ERR = [%d] \n",rc);
            break;
        }
        if (reg_cfg_tbl->mdelay_time != 0)
        {
            SEQ_LOG("i2c_write delay = %d ms\n", reg_cfg_tbl->mdelay_time);
            mdelay(reg_cfg_tbl->mdelay_time);
        }
        reg_cfg_tbl++;
    }
    SEQ_LOG("nt99250_i2c_write cnt=[%d] size [%d] \n",i,num);
    API_LOG("%s X \n", __func__);

    return rc;
}
static int32_t L_nt99250_i2c_read(
    unsigned short saddr,
    unsigned short raddr,
    unsigned short *rdata,
    enum nt99250_width_t width
)
{
    int32_t rc = 0;
    unsigned char buf[4];
    int length;
    API_LOG("%s E  slave addr =%x\n", __func__,saddr);
    if (!rdata)
    {
        ERR_LOG("%s param err NULL\n",__func__);
        return -EIO;
    }

    memset(buf, 0, sizeof(buf));


    API_LOG("%s width = [%d] \n",__func__,width);
    	    buf[0] = (raddr & 0xFF00)>>8;
    	    buf[1] = (raddr & 0x00FF);
    API_LOG("%s buf[0] = %x  buf[1]= %x\n",__func__,buf[0],buf[1]);

    switch (width)
    {
        case WORD_LEN:
            length = 2;
            break;

        case BYTE_LEN:
            length = 1;
            break;

        default:
            ERR_LOG("%s Param Err [%d]\n",__func__,rc);
            rc = -EIO;
            break;
    }

    if (!(rc < 0))
    {
        rc = L_nt99250_i2c_rxdata(saddr, buf, length);

        if (rc < 0)
        {
            ERR_LOG("%s i2c_rxdata err [%d]\n",__func__,rc);
            return rc;
        }

        switch (width)
        {
            case WORD_LEN:
                *rdata = buf[0] << 8 | buf[1];
                break;

            case BYTE_LEN:
                *rdata = buf[0];
                break;
            default:
                ERR_LOG("%s Param Err [%d]\n",__func__,rc);
                rc = -EIO;
                break;
        }
    }

    API_LOG("%s X \n", __func__);

    return rc;
}
#if 0
static int L_nt99250_i2c_read_reg_data
(
  unsigned short rpage,
  unsigned short raddr,
  unsigned short *rdata
)
{
    int rc = 0;

    unsigned short reg_val  = 0x0000;
  
    API_LOG("%s E \n", __func__);

    rc = L_nt99250_i2c_write(nt99250_client->addr, NT99250_REG_READ_PAGE, rpage, WORD_LEN);
    rc = L_nt99250_i2c_write(nt99250_client->addr, NT99250_REG_READ_ADDR, raddr, WORD_LEN);
    rc = L_nt99250_i2c_read(nt99250_client->addr, NT99250_REG_INDIRECT_DATA, &reg_val, WORD_LEN);
    if (rc < 0)
    {
        ERR_LOG("i2c_read ERR :[%d]\n",rc);
        return -EPERM;
    }

    *rdata = reg_val;
    
    SEQ_LOG("READ_REG : [0x%04x%04x] REG_DATA:[%04x]\n",rpage,raddr,*rdata);
    API_LOG("%s X \n", __func__);
    return rc;
}
#endif 
static const struct i2c_device_id nt99250_i2c_id[] = {
    { MCAM_I2C_NAME, I2C_DEVICE_ID_I2C_CAMERA},
    { },
};

static int L_nt99250_i2c_init_client
(
  struct i2c_client *client
)
{
    API_LOG("%s E \n", __func__);

    init_waitqueue_head(&nt99250_wait_queue);

    API_LOG("%s X \n", __func__);

    return 0;
}

static int L_nt99250_i2c_probe
(
  struct i2c_client *client,
  const struct i2c_device_id *id
)
{
    int rc = 0;

    API_LOG("%s E \n", __func__);

    if (!i2c_check_functionality(client->adapter, I2C_FUNC_I2C))
    {
        ERR_LOG("%s i2c_check_functionality ERR\n",__func__);
        kfree(nt99250_sensorw);
        nt99250_sensorw = NULL;
        return rc;
    }

    nt99250_sensorw = kzalloc(sizeof(struct nt99250_work), GFP_KERNEL);
    if (!nt99250_sensorw)
    {
        ERR_LOG("%s failed!\n",__func__);
        rc = -ENOMEM;
        kfree(nt99250_sensorw);
        nt99250_sensorw = NULL;
        return rc;
    }

    i2c_set_clientdata(client, nt99250_sensorw);
    L_nt99250_i2c_init_client(client);
    nt99250_client = client;

    mdelay(50);

    API_LOG("%s X \n", __func__);

    return rc;
}

static int __exit L_nt99250_i2c_remove
(
  struct i2c_client *client
)
{
    struct nt99250_work_t *sensorw = i2c_get_clientdata(client);

    API_LOG("%s E \n", __func__);

    free_irq(client->irq, sensorw);
    nt99250_client = NULL;
    nt99250_sensorw = NULL;
    kfree(sensorw);

    API_LOG("%s X \n", __func__);

    return 0;
}

static struct i2c_driver nt99250_i2c_driver = {
    .id_table = nt99250_i2c_id,
    .probe  = L_nt99250_i2c_probe,
    .remove = __exit_p(L_nt99250_i2c_remove),
    .driver = {
        .name = MCAM_I2C_NAME,
    },
};


static void L_nt99250_entry_vsync_proc
(
  uint8_t  cmd,
  uint16_t entry
)
{
    int retval;
    unsigned long flags;

    API_LOG("%s E \n", __func__);

    SEQ_LOG("%s : entry[0x%x]\n",__func__ ,entry);
    if((nt99250_ctrl->v_entry_proc==0 && entry!=0) && (nt99250_ctrl->v_int_kind != cmd))
    {
        if(nt99250_ctrl->v_int_kind != CAMERA_INT_VSYNC_NONE)
        {
            L_nt99250_delete_vsync_proc(entry);
        }

        SEQ_LOG("enable vsync irq : entry[0x%x]\n",entry);
        nt99250_ctrl->v_entry_proc |= entry;

        if(cmd == CAMERA_INT_VSYNC_HIGH)
        {
            flags = IRQF_TRIGGER_RISING;
        }
        else
        {
            flags = IRQF_TRIGGER_FALLING;
        }
        retval = request_irq(
          gpio_to_irq(nt99250_ctrl->sensordata->vsync_gpio),
          L_nt99250_vsync_interrupt, flags, "nt99250_vsync", 0);

        if( retval < 0)
        {
            ERR_LOG("[%s](%d) Err retval:[%d]\n",__func__, __LINE__,retval);
            return;
        }
        else
        {
            SEQ_LOG("%s : Interruption type [0x%x]\n",__func__ ,cmd);
            nt99250_ctrl->v_int_kind = (uint16_t)cmd;
        }
    }
    else
    {
        nt99250_ctrl->v_entry_proc |= entry;
    }

    API_LOG("%s X \n", __func__);
}

static void L_nt99250_delete_vsync_proc
(
  uint16_t entry
)
{
    API_LOG("%s E \n", __func__);
    SEQ_LOG("del entry [0x%x]\n",entry);
    nt99250_ctrl->v_entry_proc &= ~entry;

    if( (nt99250_ctrl->v_entry_proc==0)
     && ((nt99250_ctrl->v_int_kind & CAMERA_INT_VSYNC_ALWAYS) != CAMERA_INT_VSYNC_ALWAYS))
    {
        SEQ_LOG("disable vsync irq\n");
        free_irq(gpio_to_irq(nt99250_ctrl->sensordata->vsync_gpio),0);

        nt99250_ctrl->v_int_kind = CAMERA_INT_VSYNC_NONE;
    }

    API_LOG("%s X \n", __func__);
}
#if 0
static void L_nt99250_proc_get_ae_awb_status
(
  void
)
{
    int32_t rc = 0;
    static uint16_t read_ae_stable = 0;
    static uint16_t read_awb_stable = 0;
    static int v_cnt = 0;

    v_cnt++;

    API_LOG("%s E \n", __func__);

    rc = L_nt99250_i2c_read_reg_data(NT99250_REG_RAM_BASE_ADDR,CAMSENSOR_REG_AE_STABLE,&read_ae_stable);
    if( rc < 0 )
    {
        ERR_LOG("%s [ae status] I2C Read Err\n",__func__);
    }

    rc = L_nt99250_i2c_read_reg_data(NT99250_REG_RAM_BASE_ADDR,CAMSENSOR_REG_AWB_STABLE,&read_awb_stable);
    if( rc < 0 )
    {
        ERR_LOG("%s [awb status] I2C Read Err\n",__func__);
    }

    if( (read_ae_stable == 1) && (read_awb_stable == 1) )
    {
        L_nt99250_delete_vsync_proc(CAMERA_PROC_STABLE_SEARCH);
        read_ae_stable = 0;
        read_awb_stable = 0;
        L_nt99250_wake_up();
        SEQ_LOG("stable v_cnt = [%d]\n",v_cnt);
    }

    API_LOG("%s X \n", __func__);

    return;
}
static void L_nt99250_proc_read_exposure_time
(
  void
)
{
    int32_t rc = 0;
    uint16_t ET_low_data    = 0x0000;
#if 0
    uint16_t ET_upper_data  = 0x0000;
#endif
    uint32_t ET_data        = 0x00000000;
    uint16_t analog_gain    = 0x0000;

    API_LOG("%s E \n", __func__);

#if 1
    rc = L_nt99250_i2c_read_reg_data(NT99250_REG_RAM_BASE_ADDR,CAMSENSOR_REG_EXPOSURE_TIME,&ET_low_data);
    if( rc < 0 )
    {
        ERR_LOG("%s [exposure_time_L] I2C Read Err\n",__func__);
        ET_low_data = (uint16_t)EXIF_EXPOSURE_TIME_UNKNOWN;
    }
    ET_data = ((uint32_t)ET_low_data & 0x0000FFFF);
#else
    rc = L_nt99250_i2c_read_reg_data(NT99250_REG_RAM_BASE_ADDR,CAMSENSOR_REG_EXPOSURE_TIME_L,&ET_low_data);
    if( rc < 0 )
    {
        ERR_LOG("%s [exposure_time_L] I2C Read Err\n",__func__);
        ET_low_data = (uint16_t)EXIF_EXPOSURE_TIME_UNKNOWN;
    }
    rc = L_nt99250_i2c_read_reg_data(NT99250_REG_RAM_BASE_ADDR,CAMSENSOR_REG_EXPOSURE_TIME_H,&ET_upper_data);
    if( rc < 0 )
    {
        ERR_LOG("%s [exposure_time_H] I2C Read Err\n",__func__);
        ET_upper_data = (uint16_t)EXIF_EXPOSURE_TIME_UNKNOWN;
    }
    ET_data = (((uint32_t)ET_upper_data  << 16) & 0xFFFF0000) | ((uint32_t)ET_low_data & 0x0000FFFF);
#endif

    L_nt99250_i2c_read_reg_data(NT99250_REG_RAM_BASE_ADDR,CAMSENSOR_REG_ANALOG_GAIN,&analog_gain);
    nt99250_ctrl->analog_gain = analog_gain;

    DBG_LOG("%s [ET_low_data]   [%08x]\n",__func__, ET_low_data);
#if 0
    DBG_LOG("%s [ET_upper_data] [%08x]\n",__func__, ET_upper_data);
#endif
    DBG_LOG("%s [ET_data]       [%08x]\n",__func__, ET_data);

    nt99250_ctrl->exposure_time = 0;
    nt99250_ctrl->exposure_time = ((ET_data * 10) / 4);
    DBG_LOG("%s [exposure_time] [%08x]\n",__func__, nt99250_ctrl->exposure_time);
    API_LOG("%s X \n", __func__);
}
static void L_nt99250_proc_get_exposure_time
(
  void
)
{
    API_LOG("[%s] E \n",__func__);
    L_nt99250_proc_read_exposure_time();

    L_nt99250_delete_vsync_proc(CAMERA_PROC_EXPOSURE);

    API_LOG("[%s] X \n",__func__);
    return;
}
static void L_nt99250_proc_flash_on
(
  void
)
{
    int32_t rc = 0;
    static int vsync_cnt = 0;

    API_LOG("[%s] E \n",__func__);

    vsync_cnt++;

    if(vsync_cnt >= 2)
    {
        rc = L_nt99250_flash_control(CAMERA_FLASH_CTRL_MAIN);
        if( rc < 0 )
        {
            ERR_LOG("YUV Still L_nt99250_proc_flash_on Err\n");
        }

        L_nt99250_delete_vsync_proc(CAMERA_PROC_FLASH_ON);
#if 0
        L_nt99250_wake_up();
#endif
        flash_ctrl_mode = CAMERA_FLASH_CTRL_OFF;
        if( nt99250_ctrl->flash_check_result == CAMERA_FLASH_CHK_TRUE )
        {
            nt99250_ctrl->flash_check_result = CAMERA_FLASH_CHK_NONE;
        }
        SEQ_LOG("[%s] nt99250_ctrl->flash_check_result : [%d]\n",__func__ ,nt99250_ctrl->flash_check_result);
    }
    API_LOG("[%s] X \n",__func__);
    return;
}
static void L_nt99250_proc_flash_off
(
  void
)
{
    int32_t rc = 0;
    static int vsync_cnt = 0;

    API_LOG("[%s] E \n",__func__);

    vsync_cnt++;

    SEQ_LOG(" vsync_cnt:[%d]  flash_ctrl_mode:[%d]\n",vsync_cnt,flash_ctrl_mode);

    if(vsync_cnt >= 3)
    {
        L_nt99250_proc_read_exposure_time();

        rc = L_nt99250_flash_control(CAMERA_FLASH_CTRL_OFF);
        if( rc < 0 )
        {
            ERR_LOG("flash_control flash_off Err\n");
        }

        L_nt99250_delete_vsync_proc(CAMERA_PROC_FLASH_OFF);
        vsync_cnt = 0;
        flash_ctrl_mode = CAMERA_FLASH_CTRL_OFF;

        if( nt99250_ctrl->flash_check_result == CAMERA_FLASH_CHK_TRUE )
        {
            nt99250_ctrl->flash_check_result = CAMERA_FLASH_CHK_NONE;
        }
        SEQ_LOG("[%s] nt99250_ctrl->flash_check_result : [%d]\n",__func__ ,nt99250_ctrl->flash_check_result);
    }

    API_LOG("[%s] X \n",__func__);
    return;
}
static void L_nt99250_proc_flash_still
(
  void
)
{
    int32_t rc = 0;
    static int flash_light_on = 0;
    int err_flg = -1;

    API_LOG("%s E\n",__func__);
    SEQ_LOG("%s flash_ctrl_state:[%d]\n",__func__,flash_ctrl_mode);

    err_flg = 0;

    switch((int)flash_ctrl_mode)
    {
        case CAMERA_FLASH_CTRL_PRE:
            rc = L_nt99250_flash_control(CAMERA_FLASH_CTRL_PRE);
            flash_light_on = 1;
            if( rc < 0 )
            {
                err_flg = 1;
                ERR_LOG("flash_control pre_flash Err\n");
                break;
            }
            break;
        case CAMERA_FLASH_CTRL_MAIN:
            rc = L_nt99250_flash_control(CAMERA_FLASH_CTRL_MAIN);
            flash_light_on = 1;
            if( rc < 0 )
            {
                err_flg = 1;
                ERR_LOG("flash_control main_flash Err\n");
                break;
            }
            break;

        default:
            break;
    }

    L_nt99250_delete_vsync_proc(CAMERA_PROC_FLASH_STILL);
    L_nt99250_wake_up();

    if(err_flg == 1)
    {
        if(flash_light_on)
        {
            rc = L_nt99250_flash_control(CAMERA_FLASH_CTRL_OFF);
            flash_light_on = 0;
        }
    }

    API_LOG("%s X \n",__func__);

    return;
}
#endif 
#if 0
static void L_nt99250_proc_get_af_status
(
  void
)
{
    short ucAfStatus  = NT99250_AF_STATUS_IDLE;
    int rc;

    API_LOG("%s E \n",__func__);

    rc = L_nt99250_i2c_write(nt99250_client->addr, NT99250_REG_READ_PAGE, CAMSENSOR_REG_RAM_BASE_ADDR, WORD_LEN);
    if(rc < 0)
    {
        ERR_LOG("rc:[%d]  %d\n",rc, __LINE__);
    }

    rc = L_nt99250_i2c_write(nt99250_client->addr, NT99250_REG_READ_ADDR, CAMSENSOR_REG_AF_STATUS, WORD_LEN);
    if(rc < 0)
    {
        ERR_LOG("rc:[%d]  %d\n",rc, __LINE__);
    }

    rc = L_nt99250_i2c_read(nt99250_client->addr, NT99250_REG_INDIRECT_DATA, &ucAfStatus, WORD_LEN);
    if(rc < 0)
    {
        ERR_LOG("rc:[%d]  %d\n",rc, __LINE__);
    }

    SEQ_LOG("AfStatus: [%04x]  timeout_val:[%d]\n",ucAfStatus,nt99250_interrupt_count);
    if( (ucAfStatus == NT99250_AF_STATUS_SUCCSESS) || (ucAfStatus == CAMSENSOR_AF_STATUS_LOWCONF) || (ucAfStatus == CAMSENSOR_AF_STATUS_CANCEL) )
    {
        nt99250_interrupt_count = 0;
        
        L_nt99250_delete_vsync_proc(CAMERA_PROC_AF_MODE);
        
        L_nt99250_wake_up();

        SEQ_LOG("[%s](%d) ucAfStatus:[%04x] [END]\n",__func__, __LINE__,ucAfStatus);
    }
    else
    {
        nt99250_interrupt_count++ ;
        if(nt99250_interrupt_count > 72)
        {
            L_nt99250_delete_vsync_proc(CAMERA_PROC_AF_MODE);
            
            L_nt99250_wake_up();
            ERR_LOG("TIMEOUT AfStatus: [%04x]  timeout_val:[%d]\n",ucAfStatus,nt99250_interrupt_count);
            nt99250_interrupt_count = 0;
        }
        if((nt99250_interrupt_count > 10) && (ucAfStatus == NT99250_AF_STATUS_IDLE))
        {
            L_nt99250_delete_vsync_proc(CAMERA_PROC_AF_MODE);
            
            L_nt99250_wake_up();
            ERR_LOG("NO_SEARCH AfStatus: [%04x]  timeout_val:[%d]\n",ucAfStatus,nt99250_interrupt_count);
            nt99250_interrupt_count = 0;
        }
    }

    API_LOG("%s X \n",__func__);
    return;
}
static void L_nt99250_proc_regster_write_wait
(
  void
)
{
    API_LOG("%s E \n",__func__);
    L_nt99250_delete_vsync_proc(CAMERA_PROC_REGSTER_WRITE);

    L_nt99250_wake_up();

    API_LOG("%s X \n",__func__);
    return;
    
}
static void L_nt99250_proc_wait_vsync_frame
(
  void
)
{

    API_LOG("%s E \n",__func__);
    SEQ_LOG("Wait Frame val:[%d]\n",nt99250_interrupt_count);
    nt99250_interrupt_count--;

    if(nt99250_interrupt_count <= 0)
    {
        L_nt99250_delete_vsync_proc(CAMERA_PROC_WAIT_V_FRAME);
    
        L_nt99250_wake_up();

        nt99250_interrupt_count = 0;
    }

    API_LOG("%s X \n",__func__);

    return;
   
}
#endif

static void L_nt99250_proc_vsync
(
  unsigned long data
)
{
    API_LOG("%s E\n",__func__);

    if(nt99250_ctrl->v_entry_proc == CAMERA_PROC_DUMMY)
    {
        API_LOG("%s X\n",__func__);
        return;
    }

    SEQ_LOG("v_int_kind  :[0x%04x]\n",nt99250_ctrl->v_int_kind);
    SEQ_LOG("v_entry_proc:[0x%04x]\n",nt99250_ctrl->v_entry_proc);
   #if 0
    if(nt99250_ctrl->v_int_kind == CAMERA_INT_VSYNC_HIGH)
    {
        if(nt99250_ctrl->v_entry_proc&CAMERA_PROC_FLASH_STILL)
        {
            L_nt99250_proc_flash_still();
        }
        if(nt99250_ctrl->v_entry_proc&CAMERA_PROC_EXPOSURE)
        {
            L_nt99250_proc_get_exposure_time();
        }
        if(nt99250_ctrl->v_entry_proc&CAMERA_PROC_AF_MODE)
        {
          L_nt99250_proc_get_af_status();
        }
        if(nt99250_ctrl->v_entry_proc&CAMERA_PROC_REGSTER_WRITE)
        {
            L_nt99250_proc_regster_write_wait();
        }
        if(nt99250_ctrl->v_entry_proc&CAMERA_PROC_WAIT_V_FRAME)
        {
            L_nt99250_proc_wait_vsync_frame();
        }
        if(nt99250_ctrl->v_entry_proc&CAMERA_PROC_STABLE_SEARCH)
        {
            L_nt99250_proc_get_ae_awb_status();
        }
        if(nt99250_ctrl->v_entry_proc&CAMERA_PROC_CONVERGENCE_WB)
        {
            L_nt99250_proc_get_adjust_wb(st_wb_adj_info);
        }
    }
    else if(nt99250_ctrl->v_int_kind == CAMERA_INT_VSYNC_LOW)
    {
        if(nt99250_ctrl->v_entry_proc&CAMERA_PROC_FLASH_ON)
        {
            L_nt99250_proc_flash_on();
        }
        if(nt99250_ctrl->v_entry_proc&CAMERA_PROC_FLASH_OFF)
        {
            L_nt99250_proc_flash_off();
        }
    }
    else if(nt99250_ctrl->v_int_kind == CAMERA_INT_VSYNC_ALWAYS)
    {
        SEQ_LOG("%s Not Support Interruption type :[%d] \n",__func__, nt99250_ctrl->v_int_kind);
    }
    else
    {
        ERR_LOG("%s Not Support Interruption type :[%d] \n",__func__, nt99250_ctrl->v_int_kind);
    }
#endif
    API_LOG("%s X\n",__func__);
    return;
}

static irqreturn_t L_nt99250_vsync_interrupt
(
  int irq,
  void *handle
)
{
    API_LOG("%s E\n",__func__);
    L_nt99250_thread_wake_up();
    API_LOG("%s X\n",__func__);

    return IRQ_HANDLED;
}

static int L_nt99250_wait_event
(
  void
)
{
    int rc = TRUE;

    API_LOG("%s E\n",__func__);

    rc = wait_event_interruptible(nt99250_wait_queue, nt99250_ctrl->wait_flag != CAM_DRV_CURRENT_THREAD_WAIT);
    if( rc < 0 )
    {
        ERR_LOG("%s wait_event_interruptible ERR rc:[%d] \n",__func__, rc);
        rc = -1;
    }
    

    API_LOG("%s X\n",__func__);

    return rc;
}

static int L_nt99250_wake_up
(
  void
)
{
    int rc = TRUE;

    API_LOG("%s E\n",__func__);

    if(nt99250_ctrl->wait_flag == CAM_DRV_CURRENT_THREAD_WAIT)
    {
        nt99250_ctrl->wait_flag = CAM_DRV_THREAD_WAKE_UP;
        wake_up_interruptible(&nt99250_wait_queue);
        SEQ_LOG("%s  wake_up\n",__func__);
    }

    API_LOG("%s X\n",__func__);

    return rc;
}

static void L_nt99250_pwd_config
(
  unsigned on_off
)
{
    API_LOG("%s E \n", __func__);
    SEQ_LOG("[%s]:on_off=[%d]\n",__func__,on_off);

    if(on_off)
    {
#ifdef CAM_DVDD_CNT_ENABLE
        HS_A_CAM_DVDD_CNT_HI();
#else
        HS_A_CAM_VDD_D_CNT_HI();
#endif
        mdelay(1);

        HS_A_CAM_VDD_A_CNT_HI();
        mdelay(1);

        HS_A_CAM_VDD_IO_CNT_HI();
        mdelay(1);

        HS_A_CAM_VDD_AF_CNT_HI();

        mdelay(1);

        msm_camio_clk_enable(CAMIO_VFE_MDC_CLK);
        msm_camio_clk_rate_set(NT99250_DEFAULT_CLOCK_RATE);
        gpio_tlmm_config(GPIO_CFG(15, 1, GPIO_CFG_OUTPUT, GPIO_CFG_NO_PULL, GPIO_CFG_8MA), GPIO_CFG_ENABLE);

        HS_A_CAM_PWDN_HI();
        udelay(15);

        HS_A_CAM_RST_HI();
        udelay(200);

        kcj_dev_info_update_cam(DEV_INFO_CAM_PREVIEW);
    }
    else
    {   
#ifdef VREG_PULL_DOWN_SETTING
        unsigned int on_off = 1;
        unsigned int id = PM_VREG_PDOWN_GP1_ID;
#endif

        HS_A_CAM_PWDN_LO();


        HS_A_CAM_RST_LO();

        udelay(100);

        msm_camio_clk_disable(CAMIO_VFE_MDC_CLK);
        gpio_tlmm_config(GPIO_CFG(15, 1, GPIO_CFG_OUTPUT, GPIO_CFG_PULL_DOWN, GPIO_CFG_8MA), GPIO_CFG_ENABLE);
        gpio_tlmm_config(GPIO_CFG(15, 0, GPIO_CFG_OUTPUT, GPIO_CFG_PULL_DOWN, GPIO_CFG_8MA), GPIO_CFG_ENABLE);

        udelay(50);

        HS_A_CAM_VDD_AF_CNT_LO();
#ifdef VREG_PULL_DOWN_SETTING
        msm_proc_comm(PCOM_VREG_PULLDOWN, &on_off, &id);
#endif
        udelay(1500);

        HS_A_CAM_VDD_IO_CNT_LO();
#ifdef VREG_PULL_DOWN_SETTING
        id = PM_VREG_PDOWN_GP2_ID;
        msm_proc_comm(PCOM_VREG_PULLDOWN, &on_off, &id);
#endif
        udelay(1500);
#if 0
        gpio_tlmm_config(GPIO_CFG(101,  0, GPIO_INPUT, GPIO_PULL_DOWN, GPIO_8MA), GPIO_ENABLE);
        gpio_tlmm_config(GPIO_CFG(100,  0, GPIO_INPUT, GPIO_PULL_DOWN, GPIO_8MA), GPIO_ENABLE);
        mdelay(1);
#endif
        HS_A_CAM_VDD_A_CNT_LO();
#ifdef VREG_PULL_DOWN_SETTING
        id = PM_VREG_PDOWN_GP5_ID;
        msm_proc_comm(PCOM_VREG_PULLDOWN, &on_off, &id);
#endif

        udelay(1500);

#ifdef CAM_DVDD_CNT_ENABLE
        HS_A_CAM_DVDD_CNT_LO();
#else
        HS_A_CAM_VDD_D_CNT_LO();
  #ifdef VREG_PULL_DOWN_SETTING
        id = PM_VREG_PDOWN_GP6_ID;
        msm_proc_comm(PCOM_VREG_PULLDOWN, &on_off, &id);
  #endif
#endif
        mdelay(10);

        kcj_dev_info_update_cam(DEV_INFO_CAM_OFF);
    }

    API_LOG("%s X \n", __func__);

    return;
}

static int32_t L_nt99250_power_down
(
  void
)
{
    int32_t rc = 0;

    API_LOG("%s E \n", __func__);

    free_irq(gpio_to_irq(nt99250_ctrl->sensordata->vsync_gpio),0);

    L_nt99250_pwd_config(0);

    kfree(nt99250_ctrl);
    nt99250_ctrl = NULL;

    API_LOG("%s X \n", __func__);

    return rc;
}

static int32_t L_nt99250_setting
(
  enum msm_s_reg_update rupdate,
  enum msm_s_setting rsetting
)
{
    int32_t rc = 0;
    struct cam_msgbuf message;
    struct nt99250_i2c_reg_conf const* set_reg;
    uint16_t set_reg_size;

    API_LOG("%s E \n",__func__);
    SEQ_LOG("%s rupdate:[%d] rsetting:[%d]\n",__func__, rupdate ,rsetting);
    switch (rupdate)
    {
        case S_UPDATE_PERIODIC:
            if (rsetting == S_RES_PREVIEW)
            {
                set_reg = &nt99250_regs.preview_reg_settings[0];
                set_reg_size = nt99250_regs.preview_reg_settings_size;
#ifdef STILL_CONTINUANCE_FLG
                rc = 1;
#endif
            }
            else if(rsetting == S_RES_CAPTURE)
            {
                set_reg = &nt99250_regs.still_reg_settings[0];
                set_reg_size = nt99250_regs.still_reg_settings_size;
            }
            else
            {
                rc = -EINVAL;
            }
        break;

        case S_REG_INIT:
            if (rsetting == S_RES_PREVIEW || rsetting == S_RES_CAPTURE)
            {
                set_reg = &nt99250_regs.init_reg_settings[0];
                set_reg_size = nt99250_regs.init_reg_settings_size;
            }
            else
            {
                rc = -EINVAL;
            }
            break;

        default:
            rc = -EINVAL;
            break;
    }

    SEQ_LOG("%s rc:[%d]\n",__func__,rc);

    if(rc == 0)
    {
        if(rupdate != S_REG_INIT)
        {
            message.msqid    = L_nt99250_msg_get();
            message.msg_type = CAMERA_CTRL_SENSOR_MODE;
            message.msg_data = (void*)set_reg;
            message.msg_size = set_reg_size;
            message.msg_flg  = CAMERA_MSG_SYNC_MODE;

            rc = L_nt99250_msg_send(&message, CAMERA_MSG_SYNC_MODE);
        }
        else
        {
            API_LOG("nt99250_i2c_write! (rc=%d)\n",rc);
            rc = L_nt99250_i2c_write_table(set_reg, set_reg_size);
        }
        if(rc < 0)
        {
            ERR_LOG("failed nt99250_i2c_write! (rc=%d)\n",rc);
        }
        else
        {
            SEQ_LOG("nt99250_i2c_write SUCCESS (rc=%d)\n",rc);
        }
    }
    API_LOG("%s X \n",__func__);

    return rc;
}

static int32_t L_nt99250_video_config
(
  int mode,
  int res
)
{
    int32_t rc;

    API_LOG("%s E \n", __func__);
    SEQ_LOG("%s res:[%d] \n", __func__,res);
    switch (res)
    {
        case S_QTR_SIZE:
            rc = L_nt99250_setting(S_UPDATE_PERIODIC, S_RES_PREVIEW);
            if (rc < 0)
            {
                ERR_LOG("%s PREVIEW setting failed! \n", __func__);
                return rc;
            }
            break;

        case S_FULL_SIZE:
            rc = L_nt99250_setting(S_UPDATE_PERIODIC, S_RES_CAPTURE);
            if (rc < 0)
            {
                ERR_LOG("%s CAPTURE setting failed! \n", __func__);
                return rc;
            }
            break;

        default:
            ERR_LOG("%s Param Err\n", __func__);
            return 0;
    }

    nt99250_ctrl->prev_res = res;
    nt99250_ctrl->curr_res = res;
    nt99250_ctrl->sensormode = mode;

    API_LOG("%s X \n", __func__);
    return rc;
}

static int32_t L_nt99250_snapshot_config
(
  int mode
)
{
    int32_t rc = 0;

    API_LOG("%s E\n", __func__);
    SEQ_LOG("%s mode = [%d]\n", __func__, mode);

    rc = L_nt99250_setting(S_UPDATE_PERIODIC, S_RES_CAPTURE);
    if (rc < 0)
    {
        ERR_LOG("%s failed \n",__func__);
        return rc;
    }

    nt99250_ctrl->curr_res = nt99250_ctrl->pict_res;
    nt99250_ctrl->sensormode = mode;

    API_LOG("%s X\n", __func__);

    return rc;
}

static int32_t L_nt99250_raw_snapshot_config
(
  int mode
)
{
    int32_t rc = 0;

    API_LOG("%s E\n", __func__);
    SEQ_LOG("%s mode = [%d]\n", __func__, mode);

    rc = L_nt99250_setting(S_UPDATE_PERIODIC, S_RES_CAPTURE);
    if (rc < 0)
    {
        ERR_LOG("%s failed \n",__func__);
        return rc;
    }

    nt99250_ctrl->curr_res = nt99250_ctrl->pict_res;
    nt99250_ctrl->sensormode = mode;

    API_LOG("%s X  rc = %d  X\n", __func__, rc);

    return rc;
}

static int32_t L_nt99250_prepare_snapshot
(
  void
)
{
    int32_t rc = TRUE;

    API_LOG("%s E \n",__func__);

    API_LOG("%s X \n",__func__);
    return rc;
}
#if 0
static int32_t L_nt99250_start_auto_focus
(
  uint8_t focus_mode
)
{
    int32_t rc = 0;
    struct cam_msgbuf message;
    struct nt99250_i2c_reg_conf const* set_reg;
    uint16_t set_reg_size;

    API_LOG("%s E\n", __func__);
    SEQ_LOG("%s focus_mode:[%d]\n", __func__ ,focus_mode);

    rc = L_flash_ctrl_threshold_check( &nt99250_ctrl->flash_check_result );
    SEQ_LOG("[%s] nt99250_ctrl->flash_check_result : [%d]\n",__func__ ,nt99250_ctrl->flash_check_result);

    if(nt99250_ctrl->sensormode != SENSOR_PREVIEW_MODE)
    {
        ERR_LOG("%s Sensor Mode Err \n",__func__);
        return 0;
    }

    if(( nt99250_ctrl->focus_state == CAMERA_AF_FOCUS ) ||
     ( nt99250_ctrl->focus_state == CAMERA_MOVE_FOCUS ))
    {
        ERR_LOG("%s State Err \n",__func__);
        return -1;
    }

    SEQ_LOG("scene_mode (%d)\n",nt99250_ctrl->scene_mode);

    if(nt99250_ctrl->scene_mode != CAM_SCENE_AUTO)
    {
        focus_mode = scene_info[ nt99250_ctrl->scene_mode ].af_mode;
    }

    SEQ_LOG("[%s] focus_mode : [%d]\n",__func__ ,focus_mode);

    nt99250_ctrl->af_mode = focus_mode;

    if( focus_mode == CAM_AF_MODE_AUTO )
    {
        nt99250_ctrl->focus_state = CAMERA_AF_FOCUS;

        set_reg = &nt99250_regs.af_driver_data[0];
        set_reg_size = nt99250_regs.af_driver_data_size;

        message.msqid    = L_nt99250_msg_get();
        message.msg_type = CAMERA_CTRL_AF_INIT;
        message.msg_data = (void*)set_reg;
        message.msg_size = set_reg_size;
        message.msg_flg  = CAMERA_MSG_SYNC_MODE;

        rc = L_nt99250_msg_send(&message, CAMERA_MSG_SYNC_MODE);
        if (rc < 0)
        {
            ERR_LOG("%s: AF setting ERR = %d\n", __func__, rc);
            return rc;
        }
#if 0
        nt99250_ctrl->wait_flag = CAM_DRV_CURRENT_THREAD_WAIT;
        L_nt99250_entry_vsync_high_proc(CAMERA_PROC_STABLE_SEARCH);
        L_nt99250_wait_event();
#else
        mdelay(84);
#endif
    }
    else if( focus_mode == CAM_AF_MODE_MACRO )
    {
        nt99250_ctrl->focus_state = CAMERA_IDLE_FOCUS;

        nt99250_set_af_cmd[1].wdata = NT99250_REG_AF_CMD_PARAM;
        nt99250_set_af_cmd[2].wdata = 0x0078;

        message.msqid    = L_nt99250_msg_get();
        message.msg_type = CAMERA_CTRL_AF_CMD;
        message.msg_data = (void*)&nt99250_set_af_cmd[0];
        message.msg_size = ARRAY_SIZE(nt99250_set_af_cmd);
        message.msg_flg  = CAMERA_MSG_SYNC_MODE;

        rc = L_nt99250_msg_send(&message, CAMERA_MSG_SYNC_MODE);

        nt99250_interrupt_count = 1;
        nt99250_ctrl->wait_flag = CAM_DRV_CURRENT_THREAD_WAIT;
        L_nt99250_entry_vsync_proc(CAMERA_INT_VSYNC_HIGH, CAMERA_PROC_WAIT_V_FRAME);

        L_nt99250_wait_event();

        nt99250_set_af_cmd[1].wdata = NT99250_REG_AF_CMD;
        nt99250_set_af_cmd[2].wdata = NT99250_AF_MANUAL_CMD;

        message.msqid    = L_nt99250_msg_get();
        message.msg_type = CAMERA_CTRL_AF_CMD;
        message.msg_data = (void*)&nt99250_set_af_cmd[0];
        message.msg_size = ARRAY_SIZE(nt99250_set_af_cmd);
        message.msg_flg  = CAMERA_MSG_SYNC_MODE;

        rc = L_nt99250_msg_send(&message, CAMERA_MSG_SYNC_MODE);

        mdelay(17);

        nt99250_ctrl->focus_state = CAMERA_AF_FOCUS;

        set_reg = &nt99250_regs.macro_af_driver_data[0];
        set_reg_size = nt99250_regs.macro_af_driver_data_size;

        message.msqid    = L_nt99250_msg_get();
        message.msg_type = CAMERA_CTRL_AF_INIT;
        message.msg_data = (void*)set_reg;
        message.msg_size = set_reg_size;
        message.msg_flg  = CAMERA_MSG_SYNC_MODE;

        rc = L_nt99250_msg_send(&message, CAMERA_MSG_SYNC_MODE);

        if (rc < 0)
        {
            ERR_LOG("%s: AF setting ERR = %d\n", __func__, rc);
            return rc;
        }
    }

    SEQ_LOG("[%s] flash_check_result:[%d]\n",__func__ ,nt99250_ctrl->flash_check_result);

    if(nt99250_ctrl->flash_check_result == CAMERA_FLASH_CHK_TRUE)
    {
        flash_ctrl_mode = CAMERA_FLASH_CTRL_PRE;
        nt99250_ctrl->wait_flag = CAM_DRV_CURRENT_THREAD_WAIT;
        L_nt99250_entry_vsync_proc(CAMERA_INT_VSYNC_HIGH, CAMERA_PROC_FLASH_STILL);
        L_nt99250_wait_event();

        if( focus_mode == CAM_AF_MODE_NORMAL )
        {
            nt99250_interrupt_count = 10;
        }
        else
        {
            nt99250_interrupt_count = 7;
        }
        SEQ_LOG("[%s] nt99250_interrupt_count:[%d]\n",__func__ ,nt99250_interrupt_count);

        nt99250_ctrl->wait_flag = CAM_DRV_CURRENT_THREAD_WAIT;
        L_nt99250_entry_vsync_proc(CAMERA_INT_VSYNC_HIGH, CAMERA_PROC_WAIT_V_FRAME);
        L_nt99250_wait_event();
    }

    if( focus_mode != CAM_AF_MODE_NORMAL )
    {
        nt99250_set_af_cmd[1].wdata = NT99250_REG_AF_CMD;
        nt99250_set_af_cmd[2].wdata = NT99250_AF_SINGLE_CMD;

        message.msqid    = L_nt99250_msg_get();
        message.msg_type = CAMERA_CTRL_AF_CMD;
        message.msg_data = (void*)&nt99250_set_af_cmd[0];
        message.msg_size = ARRAY_SIZE(nt99250_set_af_cmd);
        message.msg_flg  = CAMERA_MSG_SYNC_MODE;

        rc = L_nt99250_msg_send(&message, CAMERA_MSG_SYNC_MODE);

        nt99250_ctrl->wait_flag = CAM_DRV_CURRENT_THREAD_WAIT;
        L_nt99250_entry_vsync_proc(CAMERA_INT_VSYNC_HIGH, CAMERA_PROC_AF_MODE);
        L_nt99250_wait_event();
    }

    nt99250_ctrl->focus_state = CAMERA_IDLE_FOCUS;

    API_LOG("%s X \n",__func__);

    return rc;
}

static int32_t L_nt99250_set_default_focus
(
  void
)
{
    int32_t rc = TRUE;
    struct cam_msgbuf message;
    struct nt99250_i2c_reg_conf const* set_reg;
    uint16_t set_reg_size;

    API_LOG("%s E \n",__func__);

    SEQ_LOG("%s flash_check_result:[%d]\n",__func__ ,nt99250_ctrl->flash_check_result);

    if( nt99250_ctrl->flash_check_result == CAMERA_FLASH_CHK_TRUE )
    {
        rc = L_nt99250_flash_control(CAMERA_FLASH_CTRL_OFF);
        if( rc < 0 )
        {
            ERR_LOG("flash_control flash_off Err\n");
        }
    }
    nt99250_ctrl->flash_check_result = CAMERA_FLASH_CHK_NONE;

    set_reg = &nt99250_regs.af_driver_data[0];
    set_reg_size = nt99250_regs.af_driver_data_size;

    message.msqid    = L_nt99250_msg_get();
    message.msg_type = CAMERA_CTRL_AF_INIT;
    message.msg_data = (void*)set_reg;
    message.msg_size = set_reg_size;
    message.msg_flg  = CAMERA_MSG_SYNC_MODE;

    rc = L_nt99250_msg_send(&message, CAMERA_MSG_SYNC_MODE);
    mdelay(84);

    API_LOG("%s X \n",__func__);
    return rc;
}

static int32_t L_nt99250_move_focus
(
  int direction,
  int32_t num_steps
)
{
    int32_t rc = TRUE;
    API_LOG("%s E \n",__func__ );
    API_LOG("%s X \n",__func__ );
    return rc;
}

static int L_nt99250_set_af_window
(
  struct afwin_info_cfg *afwin_info 
)
{
    int rc = TRUE;
    API_LOG("%s E \n",__func__ );
    API_LOG("%s X \n",__func__ );
    return rc;

}
static int L_nt99250_set_ae_window
(
  struct aeawb_win_info_cfg *afwin_info 
)
{
    int rc = TRUE;
    API_LOG("%s E \n",__func__ );
    API_LOG("%s X \n",__func__ );
    return rc;

}
static int L_nt99250_set_awb_window
(
  struct aeawb_win_info_cfg *afwin_info 
)
{
    int rc = TRUE;
    API_LOG("%s E \n",__func__ );
    API_LOG("%s X \n",__func__ );
    return rc;

}

static int L_nt99250_set_scene
(
  camera_scene_type scene_mode
)
{
    int rc = TRUE;
    struct cam_msgbuf message;
    struct nt99250_i2c_reg_conf const* set_reg;
    uint16_t set_reg_size;

    API_LOG("%s E \n",__func__ );
    SEQ_LOG("%s scene_mode:[%d]\n",__func__ ,scene_mode);

    if( scene_mode >= CAM_SCENE_MAX )
    {
        ERR_LOG("%s Param Err [%d] \n",__func__ ,scene_mode);
        return -EINVAL;
    }

    if( nt99250_ctrl->scene_mode == scene_mode )
    {
        SEQ_LOG("Same Scene Value \n");
        return TRUE;
    }

    switch( scene_mode )
    {
        case CAM_SCENE_AUTO:
            set_reg = &nt99250_regs.set_scene_auto_reg[0];
            set_reg_size = nt99250_regs.set_scene_reg_size;
            break;
        case CAM_SCENE_PORTRAIT:
            set_reg = &nt99250_regs.set_scene_portrait_reg[0];
            set_reg_size = nt99250_regs.set_scene_reg_size;
            break;
        case CAM_SCENE_SCENERY:
            set_reg = &nt99250_regs.set_scene_scenery_reg[0];
            set_reg_size = nt99250_regs.set_scene_reg_size;
            break;
        case CAM_SCENE_NIGHT_PORTRAIT:
            set_reg = &nt99250_regs.set_scene_night_portrait_reg[0];
            set_reg_size = nt99250_regs.set_scene_reg_size;
            break;
        case CAM_SCENE_NIGHT_SCENERY:
            set_reg = &nt99250_regs.set_scene_night_scenery_reg[0];
            set_reg_size = nt99250_regs.set_scene_reg_size;
            break;
        case CAM_SCENE_ACTION:
            set_reg = &nt99250_regs.set_scene_action_reg[0];
            set_reg_size = nt99250_regs.set_scene_reg_size;
            break;
        default:
            ERR_LOG("%s Param Err [%d] \n",__func__ ,scene_mode);
            return -EINVAL;
    }

    message.msqid    = L_nt99250_msg_get();
    message.msg_type = CAMERA_CTRL_SCENE;
    message.msg_data = (void*)set_reg;
    message.msg_size = set_reg_size;
    message.msg_flg  = CAMERA_MSG_SYNC_MODE;

    rc = L_nt99250_msg_send(&message, CAMERA_MSG_SYNC_MODE);
    if( rc < 0 )
    {
        ERR_LOG("failed nt99250_i2c_write! (rc=%d)\n",rc);
        return rc;
    }
    nt99250_ctrl->scene_mode = scene_mode;

    SEQ_LOG("%s nt99250_ctrl->scene_mode:[%d]\n",__func__ ,nt99250_ctrl->scene_mode);

    switch( scene_mode )
    {
        case CAM_SCENE_AUTO:
            scene_info[ CAM_SCENE_AUTO ].wb = nt99250_ctrl->wb_mode;
            scene_info[ CAM_SCENE_AUTO ].af_mode = nt99250_ctrl->af_mode;
            scene_info[ CAM_SCENE_AUTO ].ae_mode = nt99250_ctrl->aec_mode;
            scene_info[ CAM_SCENE_AUTO ].iso = nt99250_ctrl->iso_mode;
            scene_info[ CAM_SCENE_AUTO ].flash = nt99250_ctrl->flash_mode;
            scene_info[ CAM_SCENE_AUTO ].effect = nt99250_ctrl->effect_mode;
            scene_info[ CAM_SCENE_AUTO ].lens_shading = nt99250_ctrl->lens_shading_mode;
            break;
        case CAM_SCENE_PORTRAIT:
            scene_info[ CAM_SCENE_PORTRAIT ].effect = nt99250_ctrl->effect_mode;
            scene_info[ CAM_SCENE_PORTRAIT ].lens_shading = nt99250_ctrl->lens_shading_mode;
            break;
        case CAM_SCENE_SCENERY:
            scene_info[ CAM_SCENE_SCENERY ].effect = nt99250_ctrl->effect_mode;
            scene_info[ CAM_SCENE_SCENERY ].lens_shading = nt99250_ctrl->lens_shading_mode;
            break;
        case CAM_SCENE_NIGHT_PORTRAIT:
            scene_info[ CAM_SCENE_NIGHT_PORTRAIT ].effect = nt99250_ctrl->effect_mode;
            scene_info[ CAM_SCENE_NIGHT_PORTRAIT ].lens_shading = nt99250_ctrl->lens_shading_mode;
            break;
        case CAM_SCENE_NIGHT_SCENERY:
            scene_info[ CAM_SCENE_NIGHT_SCENERY ].effect = nt99250_ctrl->effect_mode;
            scene_info[ CAM_SCENE_NIGHT_SCENERY ].lens_shading = nt99250_ctrl->lens_shading_mode;
            break;
        case CAM_SCENE_ACTION:
            scene_info[ CAM_SCENE_ACTION ].effect = nt99250_ctrl->effect_mode;
            scene_info[ CAM_SCENE_ACTION ].lens_shading = nt99250_ctrl->lens_shading_mode;
            break;
        default:
            scene_info[ CAM_SCENE_AUTO ].wb = nt99250_ctrl->wb_mode;
            scene_info[ CAM_SCENE_AUTO ].af_mode = nt99250_ctrl->af_mode;
            scene_info[ CAM_SCENE_AUTO ].ae_mode = nt99250_ctrl->aec_mode;
            scene_info[ CAM_SCENE_AUTO ].iso = nt99250_ctrl->iso_mode;
            scene_info[ CAM_SCENE_AUTO ].flash = nt99250_ctrl->flash_mode;
            scene_info[ CAM_SCENE_AUTO ].effect = nt99250_ctrl->effect_mode;
            scene_info[ CAM_SCENE_AUTO ].lens_shading = nt99250_ctrl->lens_shading_mode;
            break;
    }

    API_LOG("%s X \n",__func__ );

    return rc;
}

static int L_nt99250_get_scene_info
(
  camera_scene_info_type* info
)
{
    API_LOG("%s E \n",__func__ );
    
    if( info == NULL )
    {
        ERR_LOG("%s Param NULL Err\n",__func__);
        return -EPERM;
    }

    *info = scene_info[ nt99250_ctrl->scene_mode ];
    
    API_LOG("%s X \n",__func__ );
    
    return TRUE;

}
#endif

static int L_nt99250_set_effect
(
  camera_effect_type effect
)
{
    int rc = TRUE;
    struct cam_msgbuf message;
    unsigned char write_data;

    API_LOG("%s E \n", __func__);
    SEQ_LOG("%s effect:[%d]\n",__func__ ,effect);

    if( effect >= CAM_EFFECT_MAX )
    {
        ERR_LOG("%s Param Err [%d]\n",__func__ ,effect);
        return -EFAULT;
    }

    if( nt99250_ctrl->effect_mode == effect )
    {
        SEQ_LOG("Same Effect Value \n");
        return TRUE;
    }

    switch( effect )
    {
        case CAM_EFFECT_NONE:
            write_data = NT99250_EFFECTS_NO_EFFECT;
            break;
        case CAM_EFFECT_MONO:
            write_data = NT99250_EFFECTS_MONOCHROME;
            break;
        case CAM_EFFECT_SEPIA:
            write_data = NT99250_EFFECTS_SEPIA;
            break;
        case CAM_EFFECT_NEGA:
            write_data = NT99250_EFFECTS_NEGATIVE_COLOR;
            break;
        case CAM_EFFECT_NEGA_MONO:
            write_data = NT99250_EFFECTS_NEGATIVE_MONO;
            break;
#if 0
        case CAM_EFFECT_AQUA:
            write_data = NT99250_EFFECTS_AQUA;
            break;
        case CAM_EFFECT_SKETCH:
            write_data = NT99250_EFFECTS_SKETCH;
            break;
        case CAM_EFFECT_EMBOSS_MONO:
            write_data = NT99250_EFFECTS_EMBOSS_MONO;
            break;
#endif
        default:
            ERR_LOG("%s Param Err [%d]\n",__func__ ,effect);
            return -EFAULT;
    }
    nt99250_set_effect[2].wdata = write_data;

    message.msqid    = L_nt99250_msg_get();
    message.msg_type = CAMERA_CTRL_EFFECT;
    message.msg_data = &nt99250_set_effect[0];
    message.msg_size = ARRAY_SIZE(nt99250_set_effect);
    message.msg_flg  = CAMERA_MSG_ASYNC_MODE;

    rc = L_nt99250_msg_send(&message, CAMERA_MSG_ASYNC_MODE);
    if( rc < 0 )
    {
        ERR_LOG("failed nt99250_i2c_write! (rc=%d)\n",rc);
        return FALSE;
    }

    nt99250_ctrl->effect_mode = effect;

    API_LOG("%s X \n", __func__);
    return rc;
}

#if 0
static int L_nt99250_set_wb
(
  camera_wb_type wb
)
{
    int rc = TRUE;
    int i;

    struct cam_msgbuf message;
    struct nt99250_i2c_reg_conf const* set_reg;
    uint16_t set_reg_size;

    API_LOG("%s E \n",__func__ );
    SEQ_LOG("%s wb:[%d]\n",__func__ ,wb);

    if( wb >= CAM_WB_MAX )
    {
        ERR_LOG("[%s] Param Err [%d]\n",__func__ ,wb);
        return -EINVAL;
    }

    if( nt99250_ctrl->wb_mode == wb )
    {
        SEQ_LOG("Same WB Value \n");
        return TRUE;
    }

    switch( wb )
    {
        case CAM_WB_AUTO:
            set_reg = &nt99250_regs.set_wb_auto_reg[0];
            set_reg_size = nt99250_regs.set_wb_reg_size;
            break;
        case CAM_WB_DAYLIGHT:
            set_reg = &nt99250_regs.set_wb_daylight_reg[0];
            set_reg_size = nt99250_regs.set_wb_reg_size;
            break;
        case CAM_WB_CLOUDY_DAYLIGHT:
            set_reg = &nt99250_regs.set_wb_cloudy_reg[0];
            set_reg_size = nt99250_regs.set_wb_reg_size;
            break;
        case CAM_WB_INCANDESCENT:
            set_reg = &nt99250_regs.set_wb_incandescent_reg[0];
            set_reg_size = nt99250_regs.set_wb_reg_size;
            break;
        case CAM_WB_FLUORESCENT:
            set_reg = &nt99250_regs.set_wb_fluorescent_reg[0];
            set_reg_size = nt99250_regs.set_wb_reg_size;
            break;
        default:
            ERR_LOG("%s Param Err", __func__);
            return -EINVAL;
    }

    message.msqid    = L_nt99250_msg_get();
    message.msg_type = CAMERA_CTRL_WB;
    message.msg_data = (void*)set_reg;
    message.msg_size = set_reg_size;
    message.msg_flg  = CAMERA_MSG_SYNC_MODE;

    rc = L_nt99250_msg_send(&message, CAMERA_MSG_SYNC_MODE);
    if( rc < 0 )
    {
        ERR_LOG("failed nt99250_i2c_write! (rc=%d)\n",rc);
        return rc;
    }

    if(wb != CAM_WB_AUTO &&
       nt99250_ctrl->wb_mode == CAM_WB_AUTO)
    {
        for(i=0;i!=nt99250_regs.set_wb_reg_size;i++)
        {
            if(nt99250_regs.set_wb_auto_reg[i].waddr == 0x002A &&
               nt99250_regs.set_wb_auto_reg[i].wdata == 0x0E44)
            {
                break;
            }
        }

        if(i != nt99250_regs.set_wb_reg_size)
        {
            i += 1;
            L_nt99250_i2c_read_reg_data(NT99250_REG_RAM_BASE_ADDR,0x2428,
                                              (unsigned short *)&(nt99250_regs.set_wb_auto_reg[i].wdata));
            i += 1;
            L_nt99250_i2c_read_reg_data(NT99250_REG_RAM_BASE_ADDR,0x242A,
                                              (unsigned short *)&(nt99250_regs.set_wb_auto_reg[i].wdata));
            i += 1;
            L_nt99250_i2c_read_reg_data(NT99250_REG_RAM_BASE_ADDR,0x242C,
                                              (unsigned short *)&(nt99250_regs.set_wb_auto_reg[i].wdata));

            DBG_LOG("WB Gain Take R:%04XH\n",(unsigned int)(nt99250_regs.set_wb_auto_reg[4].wdata));
            DBG_LOG("WB Gain Take G:%04XH\n",(unsigned int)(nt99250_regs.set_wb_auto_reg[5].wdata));
            DBG_LOG("WB Gain Take B:%04XH\n",(unsigned int)(nt99250_regs.set_wb_auto_reg[6].wdata));
        }
    }
    else if(wb == CAM_WB_AUTO)
    {
        memcpy((void *)nt99250_regs.set_wb_auto_reg,
               (void *)nt99250_regs.set_wb_auto_org,
               nt99250_regs.set_wb_reg_size * sizeof(struct nt99250_i2c_reg_conf));
    }

    nt99250_ctrl->wb_mode = wb;

    API_LOG("%s X \n",__func__ );

    return rc;
}

static int L_nt99250_set_iso
(
  camera_iso_type iso
)
{
    int rc = TRUE;
    struct cam_msgbuf message;
    unsigned short wdata;
    unsigned short wdata_auto;

    API_LOG("%s E \n",__func__ );
    SEQ_LOG("%s iso:[%d]\n",__func__ ,iso);

    if( iso >= CAM_ISO_MAX )
    {
        ERR_LOG("%s Param Err [%d]\n",__func__ ,iso);
        return -EINVAL;
    }

    if( nt99250_ctrl->iso_mode == iso )
    {
        SEQ_LOG("Same ISO Value \n");
        return TRUE;
    }

    wdata_auto = NT99250_ISOMODE_CLASSIC;

    switch( iso )
    {
        case CAM_ISO_AUTO:
            wdata_auto = NT99250_ISOMODE_AUTO;
            wdata = 0x0000;
            break;
        case CAM_ISO_100:
            wdata = 0x0064;
            break;
        case CAM_ISO_200:
            wdata = 0x00C8;
            break;
        case CAM_ISO_400:
            wdata = 0x0190;
            break;
        case CAM_ISO_800:
            wdata = 0x0320;
            break;
        case CAM_ISO_1600:
        default:
            ERR_LOG("%s Param Err [%d]\n",__func__,iso);
            return -EINVAL;
    }

    nt99250_set_iso[2].wdata = wdata_auto;
    nt99250_set_iso[3].wdata = wdata;

    message.msqid    = L_nt99250_msg_get();
    message.msg_type = CAMERA_CTRL_EFFECT;
    message.msg_data = &nt99250_set_effect[0];
    message.msg_size = ARRAY_SIZE(nt99250_set_effect);
    message.msg_flg  = CAMERA_MSG_ASYNC_MODE;

    rc = L_nt99250_msg_send(&message, CAMERA_MSG_ASYNC_MODE);

    nt99250_ctrl->iso_mode = iso;

    API_LOG("%s X \n",__func__ );

    return rc;
}
static int L_nt99250_set_ae_mode
(
  camera_auto_exposure_type aec_mode
)
{
    int rc = TRUE;
    struct cam_msgbuf message;
    struct nt99250_i2c_reg_conf const* set_reg;
    uint16_t set_reg_size;

    API_LOG("%s E \n",__func__ );

    SEQ_LOG("%s aec_mode:[%d]\n",__func__ ,aec_mode);

    if( aec_mode >= CAM_AEC_MAX )
    {
        ERR_LOG("%s Param Err [%d]",__func__ ,aec_mode);
        rc = -EINVAL;
        return rc;
    }
  
    if( nt99250_ctrl->aec_mode == aec_mode )
    {
        SEQ_LOG("Same AEC Value \n");
        return TRUE;
    }


    set_reg_size = nt99250_regs.set_ae_mode_reg_size;

    switch( aec_mode )
    {
        case CAM_AEC_AVERAGE:
            set_reg = &nt99250_regs.set_ae_mode_average_reg[0];
            break;
        case CAM_AEC_CENTER:
            set_reg = &nt99250_regs.set_ae_mode_center_reg[0];
            break;
        case CAM_AEC_SPOT:
            set_reg = &nt99250_regs.set_ae_mode_spot_reg[0];
            break;
        default:
            ERR_LOG("%s Param Err [%d]",__func__ ,aec_mode);
            return -EINVAL;
    }

    message.msqid    = L_nt99250_msg_get();
    message.msg_type = CAMERA_CTRL_AEC;
    message.msg_data = (void*)set_reg;
    message.msg_size = set_reg_size;
    message.msg_flg  = CAMERA_MSG_ASYNC_MODE;

    rc = L_nt99250_msg_send(&message, CAMERA_MSG_ASYNC_MODE);
    if(rc < 0)
    {
        ERR_LOG("failed nt99250_i2c_write! (rc=%d)\n",rc);
        return rc;
    }

    nt99250_ctrl->aec_mode = aec_mode;

    API_LOG("%s X \n",__func__ );

    return rc;
}

static int L_nt99250_set_antibanding
(
  int antibanding_mode
)
{
    int rc = TRUE;
    unsigned char write_data;

    API_LOG("%s E \n", __func__);

    SEQ_LOG("%s antibanding_mode:[%d]\n",__func__ ,antibanding_mode);

    if( nt99250_ctrl->antibanding_mode == antibanding_mode )
    {
        SEQ_LOG("Same antibanding Value \n");
        return TRUE;
    }

    switch( antibanding_mode )
    {
        case CAM_ANTIBANDING_50HZ:
            write_data = NT99250_FLICKE_50HZ;
            break;
        case CAM_ANTIBANDING_60HZ:
            write_data = NT99250_FLICKE_60HZ;
            break;
        case CAM_ANTIBANDING_AUTO:
            write_data = NT99250_FLICKE_AUTO;
            break;
        case CAM_ANTIBANDING_OFF:
        default:
            ERR_LOG("%s Param Err [%d] \n",__func__ ,antibanding_mode);
            return -EFAULT;
    }
#if 0
    rc = L_nt99250_i2c_write(nt99250_client->addr,
        (unsigned short)NT99250_REG_WRITE_PAGE, (unsigned short)CAMSENSOR_REG_RAM_BASE_ADDR, WORD_LEN);

    rc = L_nt99250_i2c_write(nt99250_client->addr,
        (unsigned short)NT99250_REG_WRITE_ADDR, (unsigned short)CAMSENSOR_REG_FLICKERQUANT, WORD_LEN);

    rc = L_nt99250_i2c_write(nt99250_client->addr,
        (unsigned short)NT99250_REG_INDIRECT_DATA, (unsigned short)write_data, WORD_LEN);

    if( rc < 0 )
    {
        return FALSE;
    }
#endif
    nt99250_ctrl->antibanding_mode = antibanding_mode;

    API_LOG("%s X \n", __func__);

    return rc;
}

static int L_nt99250_lens_shading
(
  uint8_t lens_shading 
)
{
    API_LOG("%s E \n",__func__ );
    SEQ_LOG("%s lens_shading:[%d]\n",__func__ ,lens_shading);

    if( nt99250_ctrl->lens_shading_mode == lens_shading )
    {
        SEQ_LOG("Same lens shading Value \n");
        return TRUE;
    }

    API_LOG("%s X \n",__func__ );

    return TRUE;
}

static int L_nt99250_set_frame_rate
(
  uint16_t  fps 
)
{
    int rc = TRUE;
    API_LOG("%s E \n",__func__ );
    SEQ_LOG("%s fps:[%d]\n",__func__ ,fps);
    API_LOG("%s X \n",__func__ );
    return rc;

}

static int L_nt99250_get_exif_info
(
  camera_exif_info *exif
)
{
    int rc = TRUE;
    camera_iso_type iso;
    uint16_t iso_speed;

    API_LOG("%s E \n", __func__);

    if(exif == NULL)
    {
        ERR_LOG("%s Param NULL Err\n",__func__);
        return -EPERM;
    }

    iso = nt99250_ctrl->iso_mode;

    if(nt99250_ctrl->scene_mode != CAM_SCENE_AUTO)
    {
      iso = scene_info[ nt99250_ctrl->scene_mode ].iso;
    }

    SEQ_LOG("get_exif_info : iso_mode      = %d\n", iso);
    switch( iso )
    {
        case CAM_ISO_AUTO:
            exif->analog_gain     = nt99250_ctrl->analog_gain;
            DBG_LOG("exif->analog_gain : [%d]\n",exif->analog_gain);
            iso_speed = CAM_ISO_AUTO;
            break;
        case CAM_ISO_100:
            iso_speed = 0x0064;
            break;
        case CAM_ISO_200:
            iso_speed = 0x00C8;
            break;
        case CAM_ISO_400:
            iso_speed = 0x0190;
            break;
        case CAM_ISO_800:
            iso_speed = 0x0320;
            break;
        case CAM_ISO_1600:
            iso_speed = 0x0640;
            break;
        default:
            ERR_LOG("%s Param Err [%d]\n",__func__ ,iso);
            iso_speed = CAM_ISO_MAX;
            break;
    }

    exif->iso_speed     = iso_speed;

    exif->exposure_time = nt99250_ctrl->exposure_time;

    exif->focal_length  = (uint16_t)EXIF_FOCAL_LENGTH;

    exif->focal_length_35mm = (uint16_t)EXIF_FOCAL_LENGTH_35MM_FILM;

    exif->f_number      = (uint16_t)EXIF_F_NUMBER;

#if 1
    exif->flash         = nt99250_ctrl->flash_on_flg;
#else
    if(nt99250_ctrl->flash_mode == CAM_FLASH_ON)
    {
        exif->flash = (EXIF_FLASH_COMP_FIRE | EXIF_FLASH_FLASH_ON);
    }
    else if(nt99250_ctrl->flash_mode == CAM_FLASH_AUTO)
    {
        exif->flash = EXIF_FLASH_AUTO_FIRE;

        if(nt99250_ctrl->flash_on_flg == TRUE)
        {
            exif->flash |= EXIF_FLASH_FLASH_ON;
        }
        else
        {
            exif->flash |= EXIF_FLASH_FLASH_OFF;
        }
    }
    else if(nt99250_ctrl->flash_mode == CAM_FLASH_OFF)
    {
        exif->flash = EXIF_FLASH_COMP_SUPPRES;
    }
    else
    {
        exif->flash = EXIF_FLASH_COMP_SUPPRES;
    }
#endif
    DBG_LOG("[EXIF_INF]: exposure_time = %d\n",  (int32_t)exif->exposure_time);
    DBG_LOG("          : iso_speed     = %d\n",  exif->iso_speed);
    DBG_LOG("          : f_number      = %d\n",  exif->f_number);
    DBG_LOG("          : focal_length  = %d\n",  exif->focal_length);
    DBG_LOG("          : focal_length_35mm = %d\n",  exif->focal_length_35mm);
    DBG_LOG("          : flash         = %d\n",  exif->flash);

    API_LOG("%s X \n", __func__);
    return rc;
}
#endif 
#if 0
static int L_nt99250_set_dimensions
(
  struct sensor_cfg_data cfg_data
)
{
    API_LOG("%s E \n",__func__ );
    API_LOG("%s X \n",__func__ );
    return TRUE;
}
#endif
static int32_t L_nt99250_set_sensor_mode
(
  int mode,
  int res
)
{
    int32_t rc = 0;

    API_LOG("%s E \n",__func__ );
    SEQ_LOG("%s mode = %d\n", __func__, mode);

    switch (mode)
    {
        case SENSOR_PREVIEW_MODE:
            if(nt99250_ctrl->sensormode == mode)
            {
                SEQ_LOG("%s Same sensormode Value %d\n", __func__, nt99250_ctrl->sensormode);
                break;
            }
            rc = L_nt99250_video_config(mode,res);
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_video_config Err\n");
            }
            break;

        case SENSOR_SNAPSHOT_MODE:
            rc = L_nt99250_snapshot_config(mode);
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_snapshot_config Err\n");
                break;
            }

#if 0
                struct nt99250_i2c_reg_conf* sensor_reg_data = NULL;
                sensor_reg_data = &nt99250_regs.set_flash_mode[0];
                sensor_reg_data[0].wdata = (long)temp_adr;
#endif

            break;

        case SENSOR_RAW_SNAPSHOT_MODE:
            rc = L_nt99250_raw_snapshot_config(mode);
            break;

        default:
            ERR_LOG("%s Param Err [%d]\n",__func__ ,mode);
            rc = -EINVAL;
            break;
    }

    if (rc < 0)
    {
        ERR_LOG("%s : ERR [%d]\n",__func__ , rc);
    }
    API_LOG("%s X \n",__func__ );
    return rc;
}

static int L_nt99250_read_hw_ver
(
  void
)
{
    int rc = 0;

    unsigned short model_id= 0x00;

    API_LOG("%s E \n",__func__ );

    rc = L_nt99250_i2c_read(nt99250_client->addr, REG_NT99250_MODEL_ID, &model_id, WORD_LEN);
   if (rc < 0)
   {
        ERR_LOG("failed nt99250_i2c_read! (rc=%d)\n",rc);
        return -EPERM;
    }


    nt99250_ctrl->hw_ver[0] = ((model_id  >> 4) & 0x0FFF);
    nt99250_ctrl->hw_ver[1] = ( model_id  & 0x000F);

    SEQ_LOG("MODEL_ID:[%02x%02x]"
        ,nt99250_ctrl->hw_ver[0],nt99250_ctrl->hw_ver[1] );

    API_LOG("%s X \n",__func__ );

    return rc;
}
#if 0
static int32_t L_nt99250_dataline_check_config
(
  int mode
)
{
    int32_t rc = 0;

    struct nt99250_i2c_reg_conf const* set_reg;
    uint16_t set_reg_size;

    API_LOG("%s E \n",__func__ );
    SEQ_LOG("%s mode:[%d] \n",__func__,mode);

    L_nt99250_pwd_config(0);

    mdelay(300);

    L_nt99250_pwd_config(1);

    set_reg = &nt99250_regs.dataline_check_mode[0];
    set_reg_size = nt99250_regs.dataline_check_mode_reg_settings_size;

    rc = L_nt99250_i2c_write_table(set_reg, set_reg_size);
    mdelay(300);

    nt99250_ctrl->curr_res = nt99250_ctrl->pict_res;
    nt99250_ctrl->sensormode = mode;

    SEQ_LOG("%s X \n",__func__);
    return rc;
}
#endif
#if 0
#ifdef FEATURE_KYOCERA_MCAM_BB_ADJ
static bool L_nt99250_set_adjust_BB_value
(
  nt99250_Blk_struct*     dat_pt,
  nt99250_ky_adjust_type  type,
  unsigned short            dat_cnt
)
{
    unsigned short i   = 0;
    long    dat_blk_No = 0;
    long    temp_adr   = 0;
    long    temp_dat   = 0;
    char    temp[4]    = {0};
    struct nt99250_i2c_reg_conf* sensor_reg_data = NULL;
    uint16_t set_reg_size = 0;

    API_LOG("%s E \n",__func__ );

    switch(type)
    {
        case NT99250_KY_ADJUST_INIT:
            sensor_reg_data = (struct nt99250_i2c_reg_conf*)&nt99250_drv_regs.init_reg_settings[0];
            set_reg_size = nt99250_regs.init_reg_settings_size;
            break;

        case NT99250_KY_ADJUST_FLASH:
            sensor_reg_data = (struct nt99250_i2c_reg_conf*)&nt99250_regs.set_flash_mode[0];
            set_reg_size = nt99250_regs.set_flash_mode_reg_size;
            break;
        case NT99250_KY_ADJUST_SCENE_AUTO:
            sensor_reg_data = (struct nt99250_i2c_reg_conf*)&nt99250_regs.set_scene_auto_reg[0];
            set_reg_size = nt99250_regs.set_scene_reg_size;
            break;
        case NT99250_KY_ADJUST_SCENE_PORTRAIT:
            sensor_reg_data = (struct nt99250_i2c_reg_conf*)&nt99250_regs.set_scene_portrait_reg[0];
            set_reg_size = nt99250_regs.set_scene_reg_size;
            break;
        case NT99250_KY_ADJUST_SCENE_SCENERY:
            sensor_reg_data = (struct nt99250_i2c_reg_conf*)&nt99250_regs.set_scene_scenery_reg[0];
            set_reg_size = nt99250_regs.set_scene_reg_size;
            break;
        case NT99250_KY_ADJUST_SCENE_NIGHT_PORTRAIT:
            sensor_reg_data = (struct nt99250_i2c_reg_conf*)&nt99250_regs.set_scene_night_portrait_reg[0];
            set_reg_size = nt99250_regs.set_scene_reg_size;
            break;
        case NT99250_KY_ADJUST_SCENE_NIGHT_SCENERY:
            sensor_reg_data = (struct nt99250_i2c_reg_conf*)&nt99250_regs.set_scene_night_scenery_reg[0];
            set_reg_size = nt99250_regs.set_scene_reg_size;
            break;
        case NT99250_KY_ADJUST_SCENE_ACTION:
            sensor_reg_data = (struct nt99250_i2c_reg_conf*)&nt99250_regs.set_scene_action_reg[0];
            set_reg_size = nt99250_regs.set_scene_reg_size;
            break;
        default :
            ERR_LOG("ERR!! Bad Set Type ");
            break;
    }

    if(dat_cnt > set_reg_size)
    {
        return TRUE;
    }

    for(i = 0; i < dat_cnt; i++)
    { 
        temp_adr   = dat_pt[i].regAdr;
        temp_dat   = dat_pt[i].regDat;
        dat_blk_No = dat_pt[i].dat_blk_No;

        DBG_LOG("-------- cnt=[%d]\n", i);
        DBG_LOG(" BlkNo:[0x%08x] Adr:[0x%08x] Data:[0x%08x]    \n",(unsigned int)dat_blk_No ,(unsigned int)temp_adr, (unsigned int)temp_dat);

        memcpy(temp,&dat_blk_No,4);

        dat_blk_No =   ((unsigned long )temp[0] << 24)
                     | ((unsigned long )temp[1] << 16)
                     | ((unsigned long )temp[2] << 8)
                     |  (unsigned long )temp[3];

        memcpy(temp,&temp_adr,4);
        temp_adr =   ((unsigned long )temp[0] << 24)
                   | ((unsigned long )temp[1] << 16)
                   | ((unsigned long )temp[2] << 8)
                   |  (unsigned long )temp[3];

        memcpy(temp,&temp_dat,4);
        temp_dat =   ((unsigned long )temp[0] << 24)
                   | ((unsigned long )temp[1] << 16)
                   | ((unsigned long )temp[2] << 8)
                   |  (unsigned long )temp[3];

        DBG_LOG(" endian -> BlkNo:[0x%08x] Adr:[0x%08x] Data:[0x%08x]    \n",(unsigned int)dat_blk_No ,(unsigned int)temp_adr, (unsigned int)temp_dat);

        dat_blk_No--;

        if(temp_adr != 0xFFFF)
        {
            sensor_reg_data[dat_blk_No].waddr = (long)temp_adr;
            sensor_reg_data[dat_blk_No].wdata = (long)temp_dat;
            DBG_LOG(" BlkNo:[0x%08x] Adr:[0x%08x] Data:[0x%08x]    \n",(unsigned int)dat_blk_No ,sensor_reg_data[dat_blk_No].waddr, (unsigned int)sensor_reg_data[dat_blk_No].wdata);
        }
        else
        {
            sensor_reg_data[dat_blk_No].waddr = (long)temp_adr;
            DBG_LOG(" Adr[0x%04x] [0x%04x]    [Skip!] \n",(unsigned int)temp_adr,(unsigned int)sensor_reg_data[dat_blk_No].wdata);
            break;
        }
    }

    API_LOG("%s X \n",__func__ );

    return TRUE;
}

static int L_nt99250_set_pict_adj
(
  char* p_data,
  unsigned long dat_blk_type,
  unsigned short data_cnt
)
{
    nt99250_Blk_struct *p_mem_work;
    int     ret;

    API_LOG("%s E \n",__func__);
    SEQ_LOG("  data_cnt:[%x]\n",data_cnt);

    p_mem_work = (nt99250_Blk_struct*)kmalloc((data_cnt * sizeof(camsensor_Blk_struct)) + 1 ,GFP_TEMPORARY);
    if(p_mem_work == NULL)
    {
        ERR_LOG("kmalloc Fault. Down");
        return -EFAULT;
    }

    if (copy_from_user((void *)p_mem_work ,(void *)p_data ,(data_cnt * sizeof(nt99250_Blk_struct))))
    {
        ERR_LOG("[%s]  EFAULT",__func__);
        kfree(p_mem_work);
        return -EFAULT;
    }

    L_nt99250_set_adjust_BB_value(p_mem_work, dat_blk_type, data_cnt);
    if(dat_blk_type == NT99250_KY_ADJUST_INIT)
    {
        ret = L_nt99250_setting(S_REG_INIT, S_RES_PREVIEW);
        ERR_LOG("[%s]  ret:[%d]",__func__ ,ret);
    }

    kfree(p_mem_work);

    API_LOG("%s X \n",__func__);

    return TRUE;
}
#endif
#endif
#if 0
static int32_t L_nt99250_wb_adjust_mode_config
(
  int mode
)
{
    int32_t rc = 0;

    struct nt99250_i2c_reg_conf const* set_reg;
    uint16_t set_reg_size;

    API_LOG("%s E \n",__func__);
    SEQ_LOG("%s mode:[%d] \n",__func__,mode);

#if 0
    L_nt99250_pwd_config(0);
    mdelay(300);
    L_nt99250_pwd_config(1);
#endif

    set_reg = &nt99250_regs.wb_adj_mode[0];
    set_reg_size = nt99250_regs.wb_adj_mode_reg_settings_size;

    rc = L_nt99250_i2c_write_table(set_reg, set_reg_size);
    mdelay(300);

    nt99250_ctrl->curr_res = nt99250_ctrl->pict_res;
    nt99250_ctrl->sensormode = mode;

    API_LOG("%s X \n",__func__);
    return rc;
}
static void L_nt99250_proc_get_adjust_wb
(
  camera_wb_adjust_info_type* wb_adj_info
)
{
    API_LOG("%s E \n",__func__ );
    SEQ_LOG("%s wb_adjust_state:[%d] \n",__func__,nt99250_ctrl->camera_wb_adjust_state);

    switch(nt99250_ctrl->camera_wb_adjust_state)
    {
      case ADJUST_WB_STATE_START:
        st_wb_adj_info = wb_adj_info;
        L_nt99250_get_adjust_wb_wait();
        break;
      case ADJUST_WB_STATE_CALC:
        L_nt99250_get_adjust_wb_calc(wb_adj_info);
        break;
      default:
        ERR_LOG("L_nt99250_proc_get_adjust_wb state err!!\n");
        break;
    }

    API_LOG("%s X \n",__func__ );
}
static void L_nt99250_get_adjust_wb_wait
(
  void
)
{
    int gain_cnt;

    API_LOG("%s E \n",__func__ );

    mdelay(NT99250_AWB_GAIN_WAIT_TIME);

    nt99250_ctrl->wb_adjust_cnt = 0;
    nt99250_ctrl->camera_wb_adjust_state = ADJUST_WB_STATE_CALC;

    for(gain_cnt=0;gain_cnt<NT99250_AWB_GAIN_READ_CNT;gain_cnt++)
    {
      nt99250_ctrl->rgain_data[gain_cnt]=0;
      nt99250_ctrl->bgain_data[gain_cnt]=0;
    }
    nt99250_ctrl->wait_flag = CAM_DRV_CURRENT_THREAD_WAIT;

    L_nt99250_entry_vsync_proc(CAMERA_INT_VSYNC_HIGH, CAMERA_PROC_CONVERGENCE_WB);

    L_nt99250_wait_event();

    API_LOG("%s X \n",__func__ );
}

static void L_nt99250_get_adjust_wb_calc
(
  camera_wb_adjust_info_type* wb_adj_info
)
{
    unsigned short wk_data;
    unsigned short R_Gain_avg    = 0x0000;
    unsigned short B_Gain_avg    = 0x0000;
    int            f_cnt;

    API_LOG("%s E \n",__func__ );

    L_nt99250_i2c_read_reg_data(NT99250_REG_RAM_BASE_ADDR, CAMSENSOR_REG_RGAIN, &wk_data);
    nt99250_ctrl->rgain_data[nt99250_ctrl->wb_adjust_cnt] = wk_data & 0xffff;
    L_nt99250_i2c_read_reg_data(NT99250_REG_RAM_BASE_ADDR, CAMSENSOR_REG_BGAIN, &wk_data);
    nt99250_ctrl->bgain_data[nt99250_ctrl->wb_adjust_cnt] = wk_data & 0xffff;

    if( nt99250_ctrl->wb_adjust_cnt == (NT99250_AWB_GAIN_READ_CNT -1) )
    {
        SEQ_LOG("avelage in E\n");
        for(f_cnt = 0; f_cnt < NT99250_AWB_GAIN_READ_CNT; f_cnt++)
        {
          R_Gain_avg += nt99250_ctrl->rgain_data[f_cnt];
          B_Gain_avg += nt99250_ctrl->bgain_data[f_cnt];
        }

        R_Gain_avg = R_Gain_avg / NT99250_AWB_GAIN_READ_CNT;
        B_Gain_avg = B_Gain_avg / NT99250_AWB_GAIN_READ_CNT;
    
        if(wb_adj_info)
        {
            wb_adj_info->rgain_min = (uint8_t)((R_Gain_avg  >> 8) & 0x00FF);
            wb_adj_info->rgain_max = (uint8_t)((R_Gain_avg      ) & 0x00FF);
            wb_adj_info->bgain_min = (uint8_t)((B_Gain_avg  >> 8) & 0x00FF);
            wb_adj_info->bgain_max = (uint8_t)((B_Gain_avg      ) & 0x00FF);

            SEQ_LOG("R_Gain_avg = [0x%x%x]", wb_adj_info->rgain_min, wb_adj_info->rgain_max );
            SEQ_LOG("B_Gain_avg = [0x%x%x]", wb_adj_info->bgain_min, wb_adj_info->bgain_max );
        }

        L_nt99250_delete_vsync_proc(CAMERA_PROC_CONVERGENCE_WB);

        L_nt99250_wake_up();
        SEQ_LOG("avelage out X\n");
    }
    else
    {
        SEQ_LOG("------------ E\n");

        nt99250_ctrl->wb_adjust_cnt++;
    }

    API_LOG("%s X \n",__func__ );
}
static int L_nt99250_get_otp_check_result
(
  int *result
)
{
    API_LOG("%s E \n",__func__);

    if( result == NULL )
    {
        ERR_LOG("%s Param NULL Err \n",__func__);
        return -EPERM;
    }

    SEQ_LOG("otp0_H [%04lx]\n",nt99250_ctrl->nt99250_otp0_H);
    SEQ_LOG("otp0_M [%04lx]\n",nt99250_ctrl->nt99250_otp0_M);
    SEQ_LOG("otp0_L [%04lx]\n",nt99250_ctrl->nt99250_otp0_L);

    do{
        *result = 0;

        if(nt99250_ctrl->nt99250_otp0_H != 0)
            break;
        if(nt99250_ctrl->nt99250_otp0_M != 0)
            break;
        if(nt99250_ctrl->nt99250_otp0_L != 0)
            break;
#if 0
        if(nt99250_ctrl->nt99250_otp1_H != 0)
            break;
        if(nt99250_ctrl->nt99250_otp1_M != 0)
            break;
        if(nt99250_ctrl->nt99250_otp1_L != 0)
            break;
#endif
        SEQ_LOG("nt99250_ctrl check result:Err");

        *result = -1;

    }while(0);

    API_LOG("%s X \n",__func__);

    return TRUE;

}

static int L_nt99250_OTP_data_read
(
  void
)
{
    int rc = -EPERM;
    int read_cnt;
    uint16_t OtpData = 0;

    API_LOG("%s E \n",__func__);

    L_nt99250_i2c_write(nt99250_client->addr, NT99250_REG_WRITE_PAGE,CAMSENSOR_REG_APB_START_ADDR, WORD_LEN);
    L_nt99250_i2c_write(nt99250_client->addr, NT99250_REG_WRITE_ADDR,CAMSENSOR_REG_CHIPID_ADDR, WORD_LEN);
    L_nt99250_i2c_write(nt99250_client->addr, NT99250_REG_INDIRECT_DATA,0x0000, WORD_LEN);

    L_nt99250_i2c_write(nt99250_client->addr, NT99250_REG_WRITE_PAGE,CAMSENSOR_REG_APB_START_ADDR, WORD_LEN);
    L_nt99250_i2c_write(nt99250_client->addr, NT99250_REG_WRITE_ADDR,CAMSENSOR_REG_CHIPID_ADDR, WORD_LEN);
    L_nt99250_i2c_write(nt99250_client->addr, NT99250_REG_INDIRECT_DATA,0x0002, WORD_LEN);

    L_nt99250_i2c_write(nt99250_client->addr, NT99250_REG_WRITE_PAGE,CAMSENSOR_REG_APB_START_ADDR, WORD_LEN);
    L_nt99250_i2c_write(nt99250_client->addr, NT99250_REG_WRITE_ADDR,CAMSENSOR_REG_CHIPID_ADDR, WORD_LEN);
    L_nt99250_i2c_write(nt99250_client->addr, NT99250_REG_INDIRECT_DATA,0x0006, WORD_LEN);

    read_cnt = 0;

    while(OtpData == 0x00)
    {
        L_nt99250_i2c_read_reg_data(NT99250_REG_APB_START_ADDR,0xA014,&OtpData);
        read_cnt++;
        if(read_cnt > 100)
        {
            break;
        }
        else
        {
            mdelay(10);
        }
    }

    if(OtpData != 0x00)
    {
        rc = L_nt99250_i2c_read_reg_data( NT99250_REG_APB_START_ADDR, 0xA00C, (unsigned short *)&nt99250_ctrl->camsensor_drv_otp0_L );
        if( rc < 0 )
        {
            ERR_LOG("%s failed! (rc=%d)\n",__func__ ,rc);
        }

        rc = L_nt99250_i2c_read_reg_data( NT99250_REG_APB_START_ADDR, 0xA00E, (unsigned short *)&nt99250_ctrl->camsensor_drv_otp0_M );
        if( rc < 0 )
        {
            ERR_LOG("%s failed! (rc=%d)\n",__func__ ,rc);
        }

        rc = L_nt99250_i2c_read_reg_data( NT99250_REG_APB_START_ADDR, 0xA010, (unsigned short *)&nt99250_ctrl->camsensor_drv_otp0_H );
        if( rc < 0 )
        {
            ERR_LOG("%s failed! (rc=%d)\n",__func__ ,rc);
        }
#if 0
        rc = L_nt99250_i2c_read_reg_data( NT99250_REG_APB_START_ADDR, 0xA00C, &nt99250_ctrl->camsensor_drv_otp1_L );
        if( rc < 0 )
        {
            ERR_LOG("%s failed! (rc=%d)\n",__func__ ,rc);
        }

        rc = L_nt99250_i2c_read_reg_data( NT99250_REG_APB_START_ADDR, 0xA00E, &nt99250_ctrl->camsensor_drv_otp1_M );
        if( rc < 0 )
        {
            ERR_LOG("%s failed! (rc=%d)\n",__func__ ,rc);
        }

        rc = L_nt99250_i2c_read_reg_data( NT99250_REG_APB_START_ADDR, 0xA010, &nt99250_ctrl->camsensor_drv_otp1_H );
        if( rc < 0 )
        {
            ERR_LOG("%s failed! (rc=%d)\n",__func__ ,rc);
        }
#endif
    }
    API_LOG("%s X \n",__func__);

    return rc;
}
#endif 
#if 0
static int L_nt99250_OTP_data_enable
(
  unsigned short enable_flg
)
{
    int rc = TRUE;

    SEQ_LOG("%s E \n",__func__);

    rc = L_nt99250_i2c_write(nt99250_client->addr, NT99250_REG_WRITE_PAGE,CAMSENSOR_REG_RAM_BASE_ADDR, WORD_LEN);
    rc = L_nt99250_i2c_write(nt99250_client->addr, NT99250_REG_WRITE_ADDR,CAMSENSOR_REG_OTP_USE_EFUSE, WORD_LEN);
    rc = L_nt99250_i2c_write(nt99250_client->addr, NT99250_REG_INDIRECT_DATA,enable_flg, WORD_LEN);

    SEQ_LOG("%s X \n",__func__);

    return rc;
}
#endif
static int L_nt99250_get_devid
(
  camera_devid_info* devid_info
)
{
    int rc = TRUE;
    uint8_t i;

    API_LOG("%s E\n",__func__);

    devid_info->ver_num = 2;

    SEQ_LOG("DEV_ID:[");
    for(i=0;i<2;i++)
    {
        devid_info->ver_area[i] = nt99250_ctrl->hw_ver[i];
        SEQ_LOG("  %02x",devid_info->ver_area[i]);
    }
    SEQ_LOG(" ]");

    devid_info->ver_check_result = TRUE;

    API_LOG("%s X\n",__func__);

    return rc;
}
static int L_nt99250_devid_check
(
  uint8_t* ver_check_result
)
{
    int rc = TRUE;

    API_LOG("%s E\n",__func__);

    *ver_check_result = CAMERA_TEST_RESULT_NG;

    do
    {
        if(nt99250_ctrl->hw_ver[0] != ((NT99250_CHIP_ID) & 0x00FF))
        {
            ERR_LOG("%s Chip_ID Err_01 [%x]\n",__func__,nt99250_ctrl->hw_ver[0]);
            break;
        }
        if(nt99250_ctrl->hw_ver[1] != (NT99250_CHIP_REV  & 0x00FF))
        {
            ERR_LOG("%s Chip_ID Err_02 [%x]\n",__func__,nt99250_ctrl->hw_ver[1]);
            break;
        }

        *ver_check_result = CAMERA_TEST_RESULT_OK;
    }while(0);

    SEQ_LOG("%s result:[%d]\n",__func__,*ver_check_result);
    API_LOG("%s X\n",__func__);

    return rc;
}

static int L_nt99250_thread_init
(
  void
)
{
    int rc = TRUE;
    int idx;

    API_LOG("%s E\n",__func__);

    init_waitqueue_head(&cam_dev_wait_queue);

    mutex_lock(&nt99250_msg_mutex);

    pst_cam_msgque = (struct cam_msgque*)kmalloc(sizeof(struct cam_msgque)*CAM_MSG_QUE_MAX,GFP_KERNEL);
    if(pst_cam_msgque == NULL)
    {
        ERR_LOG("kmalloc Fault. Down");
        return FALSE;
    }

    pst_cam_msgque->snd_idx  =  0;
    pst_cam_msgque->rcv_idx  =  0;
    pst_cam_msgque->list_num =  0;
    for(idx=0; idx<CAM_MSG_QUE_MAX; idx++)
    {
        memset(&pst_cam_msgque->pst_msg[idx],0x00,sizeof(struct cam_msgbuf));;
    }

    mutex_unlock(&nt99250_msg_mutex);

    API_LOG("%s X\n",__func__);

    return rc;
}

static void L_nt99250_thread_release
(
  void
)
{
    API_LOG("%s E\n",__func__);

    mutex_lock(&nt99250_msg_mutex);

    kfree(pst_cam_msgque);

    mutex_unlock(&nt99250_msg_mutex);

    API_LOG("%s X\n",__func__);

    return;
}

static int L_nt99250_msg_get
(
  void
)
{
    int rc = TRUE;

    API_LOG("%s E\n",__func__);

    rc = CAMERA_TYPE_FRONT;

    API_LOG("%s X\n",__func__);

    return rc;
}

static int L_nt99250_msg_send
(
  struct cam_msgbuf *msgp,
  int msgflg
)
{
    int rc = TRUE;
    int idx;

    API_LOG("%s E\n",__func__);
    SEQ_LOG("%s msgflg:[%d]\n",__func__,msgflg);

    if(msgp == NULL)
    {
        SEQ_LOG("%s Param msgp NULL\n",__func__);
        rc = FALSE;
    }
    else
    {
        mutex_lock(&nt99250_msg_mutex);

        if(pst_cam_msgque == NULL)
        {
            ERR_LOG("%s Param failed!",__func__);
            mutex_unlock(&nt99250_msg_mutex);
            return FALSE;
        }

        idx = pst_cam_msgque->snd_idx;

        DBG_LOG("%s idx [%d] \n",__func__ ,idx);
        SEQ_LOG("%s list_num [%d] \n",__func__ ,pst_cam_msgque->list_num);

        if(pst_cam_msgque->list_num >= CAM_MSG_QUE_MAX)
        {
            SEQ_LOG("%s warning Message list and full capacity!\n",__func__);

            if(msgp->msg_flg == CAMERA_MSG_SYNC_MODE)
            {
                mutex_unlock(&nt99250_msg_mutex);

                while(pst_cam_msgque->list_num >= CAM_MSG_QUE_MAX)
                {
                    mdelay(42);

                    if(pst_cam_msgque->list_num < CAM_MSG_QUE_MAX)
                    {
                        mutex_lock(&nt99250_msg_mutex);
                        break;
                    }
                }
            }
            else
            {
                ERR_LOG("%s Err Message list and full capacity!\n",__func__);
                mutex_unlock(&nt99250_msg_mutex);
                rc = FALSE;
                return rc;
            }
        }

        memcpy(&pst_cam_msgque->pst_msg[idx], msgp, sizeof(struct cam_msgbuf));

        DBG_LOG("%s msg_type [%d] \n",__func__ ,pst_cam_msgque->pst_msg[idx].msg_type);

        idx = (idx + 1) % CAM_MSG_QUE_MAX;

        pst_cam_msgque->snd_idx = idx;
        pst_cam_msgque->list_num++;

        DBG_LOG("%s snd_idx [%d] \n",__func__ ,pst_cam_msgque->snd_idx);
        DBG_LOG("%s list_num [%d] \n",__func__ ,pst_cam_msgque->list_num);

        mutex_unlock(&nt99250_msg_mutex);

        SEQ_LOG("%s msgp->msg_flg[%d] \n",__func__,msgp->msg_flg);

        if(msgp->msg_flg == CAMERA_MSG_SYNC_MODE)
        {
            nt99250_ctrl->wait_flag = CAM_DRV_CURRENT_THREAD_WAIT;
            L_nt99250_wait_event();
        }

    }

    API_LOG("%s X\n",__func__);

    return rc;
}

static camera_msg_err_code L_nt99250_msg_rcv
(
  struct cam_msgbuf *msgp,
  int msgtyp,
  int msgflg
)
{
    int ret;
    camera_msg_err_code err_code = CAMERA_MSG_CTRL_INIT;
    int idx = 0;

    API_LOG("%s E\n",__func__);

    if(msgp == NULL)
    {
        ERR_LOG("%s Param msgp NULL",__func__);
        return FALSE;
    }

    nt99250_ctrl->wait_thread_flag = CAM_DRV_CURRENT_THREAD_WAIT;
    ret = L_nt99250_thread_wait_event();

    if (L_nt99250_thread_should_stop())
    {
        SEQ_LOG("%s Process end by Deactivate request\n",__func__);
        err_code = CAMERA_MSG_CTRL_OTHER_ERR;
        return err_code;
    }
    mutex_lock(&nt99250_msg_mutex);

    if(pst_cam_msgque == NULL)
    {
        ERR_LOG("%s Param pst_cam_msgque NULL\n",__func__);
        err_code = CAMERA_MSG_CTRL_FAIL;
        mutex_unlock(&nt99250_msg_mutex);
        return err_code;
    }

    idx = pst_cam_msgque->rcv_idx;

    if(idx >= CAM_MSG_QUE_MAX)
    {
        ERR_LOG("%s Param Index Outside range \n",__func__);
        err_code = CAMERA_MSG_CTRL_FAIL;
        mutex_unlock(&nt99250_msg_mutex);
        return err_code;
    }

    if(( pst_cam_msgque->pst_msg[idx].msg_type > CAMERA_CTRL_INIT )
     &&( pst_cam_msgque->pst_msg[idx].msg_type < CAMERA_CTRL_MAX  ))
    {

        SEQ_LOG("%s msgtyp:[%d]\n",__func__ ,pst_cam_msgque->pst_msg[idx].msg_type);

        DBG_LOG("%s rcv_idx [%d] \n",__func__ ,pst_cam_msgque->rcv_idx);
        DBG_LOG("%s idx [%d] \n",__func__ ,idx);

        memcpy(msgp, &pst_cam_msgque->pst_msg[idx], sizeof(struct cam_msgbuf));

        memset(&pst_cam_msgque->pst_msg[idx],0x00,sizeof(struct cam_msgbuf));

        pst_cam_msgque->rcv_idx = ((idx + 1) % CAM_MSG_QUE_MAX);

        pst_cam_msgque->list_num--;
        DBG_LOG("%s rcv_idx [%d] \n",__func__ ,pst_cam_msgque->rcv_idx);
        DBG_LOG("%s list_num [%d] \n",__func__ ,pst_cam_msgque->list_num);
        err_code = CAMERA_MSG_CTRL_SUCCESS;
    }
    else
    {
        err_code = CAMERA_MSG_CTRL_NO_MSG;
    }

    mutex_unlock(&nt99250_msg_mutex);

    API_LOG("%s X\n",__func__);

    return err_code;
}

static int L_nt99250_msg_proc
(
  struct cam_msgbuf *msgp
)
{
    int rc = TRUE;

    API_LOG("%s E\n",__func__);

    if(msgp == NULL)
    {
        ERR_LOG("%s Param msgp NULL",__func__);
        return FALSE;
    }

    if(msgp->msg_type == CAMERA_CTRL_INIT)
    {
        SEQ_LOG("%s No Message",__func__);
        return TRUE;
    }

    SEQ_LOG("%s msg_type[%d] msg_flg[%d]\n",__func__ ,msgp->msg_type ,msgp->msg_flg);

    if((msgp->msg_type > CAMERA_CTRL_INIT)
     &&(msgp->msg_type < CAMERA_CTRL_MAX ))
    {

        rc = L_nt99250_register_accsess((camera_cmd_type)msgp->msg_type, msgp->msg_data, msgp->msg_size);
        msgp->msg_type = CAMERA_CTRL_INIT;

        if(msgp->msg_flg == CAMERA_MSG_SYNC_MODE)
        {
            L_nt99250_wake_up();
        }
    }
    else
    {
        rc = FALSE;
        ERR_LOG("%s Param msg_type Err\n",__func__);
    }

    API_LOG("%s X\n",__func__);
    return rc;
}

static int L_nt99250_register_accsess
(
  camera_cmd_type cmd,
  void* data,
  uint16_t size
)
{
    int rc = TRUE;
    struct nt99250_i2c_reg_conf const* reg_data;
    uint16_t set_data_size;

    API_LOG("%s E\n",__func__);

    SEQ_LOG("%s  cmd (%d)\n",__func__ ,cmd);
    SEQ_LOG("%s  data(0x%08x)\n",__func__ ,(unsigned int)data);
    SEQ_LOG("%s  size(%d)\n",__func__ ,size);


    switch((int)cmd)
    {
        case CAMERA_CTRL_SENSOR_MODE:
        case CAMERA_CTRL_AF_INIT:
        case CAMERA_CTRL_AF_CMD:
        case CAMERA_CTRL_WB:
        case CAMERA_CTRL_EFFECT:
        case CAMERA_CTRL_AEC:
        case CAMERA_CTRL_SCENE:
            reg_data = data;
            set_data_size = size;
            break;

        case CAMERA_CTRL_ISO:
        case CAMERA_CTRL_ANTIBANDING:
        case CAMERA_CTRL_CONTRAST:
        case CAMERA_CTRL_BRIGHTNESS:
        case CAMERA_CTRL_SHARPNESS:
        case CAMERA_CTRL_LENS_SHADING:
        case CAMERA_CTRL_FLASH:
        case CAMERA_CTRL_MAX:
        default:
            reg_data = 0;
            set_data_size = 0;
            SEQ_LOG("%s Not Support Ctrl \n",__func__);
            break;
    }

    if((reg_data == NULL) || (set_data_size == 0))
    {
        rc = FALSE;
        ERR_LOG("%s Param NULL Err\n",__func__);
    }
    else
    {
        rc = L_nt99250_i2c_write_table(reg_data, set_data_size);
    }

    API_LOG("%s X\n",__func__);

    return rc;
}

static int L_nt99250_thread_wait_event
(
  void
)
{
    int rc ;

    API_LOG("%s E\n",__func__);

    rc = wait_event_interruptible(cam_dev_wait_queue, nt99250_ctrl->wait_thread_flag != CAM_DRV_CURRENT_THREAD_WAIT);
    if( rc < 0 )
    {
        ERR_LOG("%s Err\n",__func__);
        rc = -1;
    }

    API_LOG("%s X\n",__func__);

    return rc;
}

static int L_nt99250_thread_wake_up
(
  void
)
{
    int rc = TRUE;

    API_LOG("%s E\n",__func__);

    if(nt99250_ctrl->v_entry_proc != 0)
    {
        if(nt99250_ctrl->wait_thread_flag == CAM_DRV_CURRENT_THREAD_WAIT)
        {
            nt99250_ctrl->wait_thread_flag = CAM_DRV_THREAD_WAKE_UP;
            wake_up_interruptible(&cam_dev_wait_queue);
            SEQ_LOG("%s  kthread wake_up\n",__func__);
        }
    }
    else
    {
        SEQ_LOG("%s  No Entry Process \n",__func__);
    }

    API_LOG("%s X\n",__func__);

    return rc;
}

static int L_nt99250_thread_run
(
  void
)
{
    int rc = TRUE;

    API_LOG("%s E\n",__func__);

    if(L_nt99250_thread_init() != TRUE)
    {
        ERR_LOG("%s Initialization failure\n",__func__);
        return FALSE;
    }

    nt99250_ctrl->camdev_th = kthread_run(L_nt99250_thread, &nt99250_ctrl->sensordata, "camdev_thread");

    nt99250_ctrl->v_int_kind = CAMERA_INT_VSYNC_NONE;

    if (IS_ERR(nt99250_ctrl->camdev_th))
    {
        ERR_LOG("kthread create failed!\n");
        rc = PTR_ERR(nt99250_ctrl->camdev_th);
        nt99250_ctrl->camdev_th = NULL;
    }
    
    API_LOG("%s X\n",__func__);

    return rc;
}

static int L_nt99250_thread_stop
(
  void
)
{
    int rc = TRUE;

    API_LOG("%s E\n",__func__);

    rc = kthread_stop(nt99250_ctrl->camdev_th);

    L_nt99250_thread_release();

    API_LOG("%s X\n",__func__);

    return rc;
}

static int L_nt99250_thread_should_stop
(
  void
)
{
    int rc = TRUE;

    API_LOG("%s E\n",__func__);

    rc = kthread_should_stop();
    
    API_LOG("%s X\n",__func__);

    return rc;
}

static int L_nt99250_thread
(
  void *data
)
{
    int rc;
    camera_msg_err_code err_code;
    struct cam_msgbuf *msgp;
    unsigned long dummy = 0;

    API_LOG("%s E \n", __func__);

    msgp = (struct cam_msgbuf*)kmalloc(sizeof(struct cam_msgbuf),GFP_KERNEL);

    memset(msgp, 0x00, sizeof(struct cam_msgbuf));


    while (!L_nt99250_thread_should_stop())
    {
        if(msgp == NULL)
        {
            ERR_LOG("%s The buffer for the message reception is NULL \n",__func__);
            break;
        }

        err_code = L_nt99250_msg_rcv(msgp, 0,0);
        if(err_code == CAMERA_MSG_CTRL_FAIL)
        {
            ERR_LOG("%s Process end by error, err_code:[%d]\n",__func__,err_code);
            break;
        }
        else if(err_code == CAMERA_MSG_CTRL_OTHER_ERR)
        {
            SEQ_LOG("%s Process end by Deactivate request\n",__func__);
            break;
        }

        L_nt99250_proc_vsync(dummy);

        if(err_code != CAMERA_MSG_CTRL_NO_MSG)
        {
            SEQ_LOG("%s The message demand processing is executed \n",__func__);
            rc = L_nt99250_msg_proc(msgp);
            if( rc < 0 )
            {
                SEQ_LOG("%s Message Processing failure :[%d]\n",__func__,rc);
                break;
            }
        }

        schedule();
    }

    kfree(msgp);

    SEQ_LOG("%s Process end \n",__func__);
    API_LOG("%s X \n", __func__);
    return 0;
}

#ifdef DBG_TIME_LOG_SWITCH
static void L_nt99250_dbg_proc_time
(
  dbg_time_cmd cmd_type
)
{
    unsigned int time;
    static struct timeval start_time;
    static struct timeval stop_time;

    API_LOG("%s E \n", __func__);

    if(cmd_type == DBG_TIME_INIT)
    {
        start_time.tv_sec  = 0;
        start_time.tv_usec = 0;
        stop_time.tv_sec   = 0;
        stop_time.tv_usec  = 0;
    }
    else
    {
        if(cmd_type == DBG_TIME_START)
        {
            do_gettimeofday(&start_time);
        }
        else if(cmd_type == DBG_TIME_END)
        {
            do_gettimeofday(&stop_time);

            time = stop_time.tv_sec - start_time.tv_sec;
            time *= 1000000;
            if (stop_time.tv_usec < start_time.tv_usec)
            {
                time -= start_time.tv_usec - stop_time.tv_usec;
            }
            else
            {
                time += stop_time.tv_usec - start_time.tv_usec;
            }

            DBG_TIME_LOG("Processing time = [%d]usec\n", time);

            start_time.tv_sec  = 0;
            start_time.tv_usec = 0;
            stop_time.tv_sec   = 0;
            stop_time.tv_usec  = 0;
        }
        else
        {
            DBG_TIME_LOG("Err Parameter of non-support [%d]\n",cmd_type);
        }

    }
    API_LOG("%s X \n", __func__);
    return;
}
#endif

static int L_nt99250_reset(const struct msm_camera_sensor_info *dev)
{
        int rc = 0;
        API_LOG("%s S \n", __func__);

        rc = gpio_request(dev->sensor_reset, "nt99250");

        if (!rc) {
                rc = gpio_direction_output(dev->sensor_reset, 0);
                mdelay(20);
                rc = gpio_direction_output(dev->sensor_reset, 1);
        }

        gpio_free(dev->sensor_reset);
        API_LOG("%s X \n", __func__);
        return rc;
}


static int L_nt99250_probe_init_sensor
(
  const struct msm_camera_sensor_info *data
)
{
    int rc = 0;
    API_LOG("%s E \n",__func__ );
     
    rc = L_nt99250_reset(data);
    if (rc < 0) {
            ERR_LOG("reset failed!\n");
            return rc;
        }
     mdelay(5);
    API_LOG("%s X \n",__func__ );
    return rc;
}

static int L_nt99250_set_reg_table
(
  void
)
{
    int rc = 0;
    API_LOG("%s E \n",__func__ );

    nt99250_ctrl->nt99250_regs = &nt99250_regs;

    API_LOG("%s X \n",__func__ );

    return rc;
}

static int L_nt99250_sensor_init
(
  const struct msm_camera_sensor_info *data
)
{
    int32_t  rc;

    API_LOG("%s E \n",__func__);

    nt99250_ctrl = kzalloc(sizeof(struct nt99250_ctrl), GFP_KERNEL);
    if (!nt99250_ctrl)
    {
        ERR_LOG("nt99250_init failed!\n");
        rc = -ENOMEM;
        return rc;
    }
    nt99250_ctrl->set_test = S_TEST_OFF;
    nt99250_ctrl->prev_res = S_QTR_SIZE;
    nt99250_ctrl->pict_res = S_FULL_SIZE;

    nt99250_ctrl->flash_mode   = CAM_FLASH_OFF;

    nt99250_ctrl->wb_mode      = CAM_WB_AUTO;
    nt99250_ctrl->effect_mode  = CAM_EFFECT_NONE;
    nt99250_ctrl->aec_mode     = CAM_AEC_AVERAGE;


    nt99250_ctrl->camera_wb_adjust_state = ADJUST_WB_STATE_NONE;
    if (data)
    {
        nt99250_ctrl->sensordata = data;
    }

    nt99250_ctrl->wait_flag           = CAM_DRV_THREAD_WAKE_UP;
    nt99250_ctrl->nt99250_regs       = NULL;

    nt99250_ctrl->flash_on_flg = 0;

    nt99250_ctrl->nt99250_otp0_H = 0;
    nt99250_ctrl->nt99250_otp0_M = 0;
    nt99250_ctrl->nt99250_otp0_L = 0;
    flash_ctrl_mode = CAMERA_FLASH_CTRL_OFF;

    L_nt99250_pwd_config(1);

    msm_camio_camif_pad_reg_reset();
    mdelay(20);

    rc = L_nt99250_read_hw_ver();
    if (rc < 0)
    {
        ERR_LOG("L_nt99250_read_hw_ver failed!\n");
        L_nt99250_power_down();
        return rc;
    }

    rc = L_nt99250_set_reg_table();
    if (rc < 0)
    {
        ERR_LOG("L_nt99250_set_reg_table failed!\n");
        L_nt99250_power_down();
        return rc;
    }

#if 0
    rc = L_nt99250_probe_init_sensor(data);
    if (rc < 0)
    {
        goto init_fail;
    }
#endif

    SEQ_LOG("%s prev_res:[%d]\n",__func__ ,nt99250_ctrl->prev_res);
    if (nt99250_ctrl->prev_res == S_QTR_SIZE)
    {
        rc = L_nt99250_setting(S_REG_INIT, S_RES_PREVIEW);
        nt99250_ctrl->sensormode = SENSOR_PREVIEW_MODE;
    }
    else
    {
        rc = L_nt99250_setting(S_REG_INIT, S_RES_CAPTURE);
        nt99250_ctrl->sensormode = SENSOR_SNAPSHOT_MODE;
    }
 
    SEQ_LOG("%s rc:[%d]\n",__func__ ,rc);
    if (rc < 0)
    {
        ERR_LOG("L_nt99250_setting failed. rc = %d\n", rc);
        L_nt99250_power_down();
    }
    else
    {
        rc = L_nt99250_thread_run();
        if (rc < 0)
        {
            ERR_LOG("L_nt99250_thread_run failed. rc = %d\n", rc);
            L_nt99250_thread_release();
            L_nt99250_power_down();
        }
        L_nt99250_entry_vsync_proc(CAMERA_INT_VSYNC_HIGH, CAMERA_PROC_DUMMY);
    }

    SEQ_LOG("L_nt99250_sensor_init (rc=%d)",rc);
    API_LOG("%s X \n",__func__);
    return rc;
}

static int L_nt99250_sensor_release
(
  void
)
{
    int rc = -EBADF;

    API_LOG("%s E \n",__func__);

    mutex_lock(&nt99250_mutex);

    L_nt99250_thread_stop();

    L_nt99250_power_down();

    SEQ_LOG("nt99250_release completed\n");
    mutex_unlock(&nt99250_mutex);

    API_LOG("%s X \n",__func__);

    return rc;
}

static int L_nt99250_sensor_config
(
  void __user *argp
)
{
    struct sensor_cfg_data cdata;
    int32_t rc = 0;

    API_LOG("%s E \n",__func__);

    if (copy_from_user(&cdata ,(void *)argp ,sizeof(struct sensor_cfg_data)))
    {
        ERR_LOG("%s Err\n",__func__);
        return -EFAULT;
    }

    mutex_lock(&nt99250_mutex);

    SEQ_LOG("%s: cfgtype = %d\n", __func__, cdata.cfgtype);
    switch (cdata.cfgtype)
    {
        case CFG_SET_MODE:
            rc = L_nt99250_set_sensor_mode(cdata.mode, cdata.rs);
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_set_sensor_mode Err :[%d]\n",rc);
            }
            break;
#if 0
        case CFG_MOVE_FOCUS:
            rc = L_nt99250_move_focus(cdata.cfg.focus.dir,cdata.cfg.focus.steps);
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_move_focus Err :[%d]\n",rc);
            }
            break;

        case CFG_SET_DEFAULT_FOCUS:
            rc = L_nt99250_set_default_focus();
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_set_default_focus Err :[%d]\n",rc);
            }
            break;

        case CFG_SET_AUTO_FOCUS:
            rc = L_nt99250_start_auto_focus((uint8_t)cdata.mode);
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_start_auto_focus Err :[%d]\n",rc);
            }
            break;
        case CFG_SET_WB:
            rc = L_nt99250_set_wb((camera_wb_type)cdata.cfg.value );
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_set_wb Err :[%d]\n",rc);
            }
            break;
#endif
        case CFG_SET_EFFECT:
            rc = L_nt99250_set_effect((camera_effect_type)cdata.cfg.value );
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_set_effect Err :[%d]\n",rc);
            }
            break;
#if 0
        case CFG_SET_EXPOSURE_MODE:
            rc = L_nt99250_set_ae_mode((camera_effect_type)cdata.cfg.value );
            if( rc < 0 )
            {
                ERR_LOG("nt99250_set_ae_mode Err :[%d]\n",rc);
            }
            break;

        case CFG_SET_SCENE:
            rc = L_nt99250_set_scene(cdata.mode);
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_set_scene Err :[%d]\n",rc);
            }
            break;

        case CFG_GET_SCENE_INFO:
            rc = L_nt99250_get_scene_info(&(cdata.scene_info ));
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_get_scene_info Err :[%d]\n",rc);
            }
            else
            {
                if( copy_to_user( argp, &cdata, sizeof( struct sensor_cfg_data )))
                {
                    ERR_LOG("CFG_GET_SCENE_INFO copy_to_user() Err\n");
                    rc = -EPERM;
                }
                else
                {
                    rc = TRUE;
                }
            }
            break;
        case CFG_GET_WB_ADJUST_VALUE:
            nt99250_ctrl->camera_wb_adjust_state = ADJUST_WB_STATE_START;
            L_nt99250_proc_get_adjust_wb(&(cdata.wb_adj_info));

            if( copy_to_user( argp, &cdata, sizeof( struct sensor_cfg_data )))
            {
                ERR_LOG("CFG_GET_WB_ADJUST_VALUE copy_to_user() Err\n");
                rc = -EPERM;
            }
            else
            {
                rc = TRUE;
            }
            break;
        case CFG_SET_ANTIBANDING:
            rc = L_nt99250_set_antibanding( cdata.mode );
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_set_antibanding Err :[%d]\n",rc);
            }
            break;

        case CFG_SET_DIMENSIONS:
            rc = L_nt99250_set_dimensions( cdata );
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_set_dimensions Err :[%d]\n",rc);
            }
            break;
#if 0

#ifdef FEATURE_KYOCERA_MCAM_BB_ADJ
        case CFG_DEBUG_SET_PICT_ADJ:
            rc = L_nt99250_set_pict_adj( (char *)cdata.BB_Adjust_data, 
                                           cdata.BB_Adjust_data_type,
                                           cdata.BB_Adjust_data_size);
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_set_pict_adj Err :[%d]\n",rc);
            }
            break;
#endif
        case CFG_SET_ISO:
            rc = L_nt99250_set_iso(cdata.cfg.value );
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_set_iso Err :[%d]\n",rc);
            }
            break;
#endif
        case CFG_SET_LENS_SHADING:
            rc = L_nt99250_lens_shading(( uint8_t )cdata.cfg.lens_shading);
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_lens_shading Err :[%d]\n",rc);
            }
            break;

        case CFG_GET_EXIF_INFO:
            rc = L_nt99250_get_exif_info(&(cdata.exif_info));
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_get_exif_info Err :[%d]\n",rc);
            }
            else
            {
                if( copy_to_user( argp, &cdata, sizeof( struct sensor_cfg_data )))
                {
                    ERR_LOG("CFG_GET_EXIF_INFO copy_to_user() Err\n");
                    rc = -EPERM;
                }
                else
                {
                    rc = TRUE;
                }
            }
            break;
        case CFG_GET_OTP_CHECK_RESULT:
            rc = L_nt99250_OTP_data_read();
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_OTP_data_read Err :[%d]\n",rc);
            }
            rc = L_nt99250_get_otp_check_result(&(cdata.otp_check_result));
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_get_otp_check_result Err :[%d]\n",rc);
            }
            else
            {
                if( copy_to_user( argp, &cdata, sizeof(struct sensor_cfg_data)))
                {
                    ERR_LOG("L_nt99250_get_otp_check_result copy_to_user() Err\n");
                    rc = -EPERM;
                }
                else
                {
                    rc = TRUE;
                }
            }
            break;
#endif
        case CFG_PREPARE_SNAPSHOT:
            rc = L_nt99250_prepare_snapshot();
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_prepare_snapshot Err :[%d]\n",rc);
            }
            break;
#if 0
        case CFG_SET_AF_WINDOW:
            rc = L_nt99250_set_af_window(&(cdata.cfg.afwin_info));
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_set_af_window Err :[%d]\n",rc);
            }
            break;
        case CFG_SET_AE_WINDOW:
            rc = L_nt99250_set_ae_window(&(cdata.cfg.aewin_info));
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_set_ae_window Err :[%d]\n",rc);
            }
            break;

        case CFG_SET_AWB_WINDOW:
            rc = L_nt99250_set_awb_window(&(cdata.cfg.awbwin_info));
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_set_awb_window Err :[%d]\n",rc);
            }
            break;
#endif
  #if 0

        case CFG_SET_FPS:
            rc = L_nt99250_set_frame_rate((uint16_t)cdata.cfg.gfps.prevfps );
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_set_frame_rate Err :[%d]\n",rc);
            }
            break;
        case CFG_SET_CONTRAST:
            rc = nt99250_set_contrast(cdata.mode);
            if( rc < 0 )
            {
                ERR_LOG("nt99250_set_contrast Err :[%d]\n",rc);
            }
            break;

  #endif
  #if 0
        case CFG_SET_BRIGHTNESS:
            rc = nt99250_set_brightness(cdata.mode);
            if( rc < 0 )
            {
                ERR_LOG("nt99250_set_brightness Err :[%d]\n",rc);
            }
            break;

  #endif
  #if 0
        case CFG_SET_SHARPNESS:
            rc = nt99250_set_sharpness(cdata.mode);
            if( rc < 0 )
            {
                ERR_LOG("nt99250_set_sharpness Err :[%d]\n",rc);
            }
           rc = TRUE;

            break;
  #endif
#if 0
        case CFG_SET_STILL_TO_PREVIEW:
            rc = nt99250_still_to_preview();
            if( rc < 0 )
            {
                ERR_LOG("nt99250_still_to_preview Err :[%d]\n",rc);
            }
            break;
#endif
        case CFG_GET_DEVID:
            rc = L_nt99250_get_devid(&(cdata.devid_info));
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_get_devid Err :[%d]\n",rc);
                break;
            }

            if( copy_to_user( argp, &cdata, sizeof(struct sensor_cfg_data)))
            {
                ERR_LOG("CFG_GET_DEVID copy_to_user() Err\n");
                rc = -EPERM;
            }
            else
            {
                rc = TRUE;
            }
            break;

        case CFG_DEVID_CHECK:
            rc = L_nt99250_devid_check(&(cdata.devid_info.ver_check_result));
            if( rc < 0 )
            {
                ERR_LOG("L_nt99250_devid_check Err :[%d]\n",rc);
                break;
            }

            if( copy_to_user( argp, &cdata, sizeof(struct sensor_cfg_data)))
            {
                ERR_LOG("L_cam_kernel_get_otp_check_result copy_to_user() Err\n");
                rc = -EPERM;
            }
            else
            {
                rc = TRUE;
            }
            break;

        default:
            ERR_LOG("%s Param Err [%d]\n",__func__ ,cdata.cfgtype);
            rc = -EINVAL;
            break;
    }

    mutex_unlock(&nt99250_mutex);

    API_LOG("%s X \n",__func__);

    return rc;
}

static int L_nt99250_sensor_probe
(
  const struct msm_camera_sensor_info *info,
  struct msm_sensor_ctrl *sensor
)
{
    int rc = 0;

    API_LOG("%s E \n",__func__);

    rc = i2c_add_driver(&nt99250_i2c_driver);
    if (rc < 0)
    {
        ERR_LOG("failed nt99250 i2c_add_driver! (rc=%d)\n", rc);
        ERR_LOG("SENSOR PROBE FAILS!\n");
        rc = -ENOTSUPP;
        return rc;
    }

#ifdef I2C_DEVICE_AS3685C
    rc = i2c_add_driver(&flash_ctrl_i2c_driver);
    if (rc < 0)
    {
        ERR_LOG("failed flash i2c_add_driver! (rc=%d)\n", rc);
        rc = -ENOTSUPP;
    }
#endif

    rc = L_nt99250_probe_init_sensor(info);
    if (rc < 0)
    {
        ERR_LOG("SENSOR PROBE FAILS!\n");
        return rc;
    }

    sensor->s_init = L_nt99250_sensor_init;
    sensor->s_release = L_nt99250_sensor_release;
    sensor->s_config  = L_nt99250_sensor_config;
    sensor->s_mount_angle  = 0;

    API_LOG("%s X \n",__func__);

    return rc;
}

static int L_nt99250_probe
(
  struct platform_device *pdev
)
{
    return msm_camera_drv_start(pdev, L_nt99250_sensor_probe);
}

static struct platform_driver msm_camera_driver = {
    .probe = L_nt99250_probe,
    .driver = {
        .name = MCAM_SENSOR_DEVICE_NAME,
        .owner = THIS_MODULE,
    },
};

static int __init nt99250_init
(
  void
)
{
    return platform_driver_register(&msm_camera_driver);
}

module_init(nt99250_init);

