/* This software is contributed or developed by KYOCERA Corporation. */
/* (C) 2012 KYOCERA Corporation                                      */

#include <asm/div64.h>
#include <linux/string.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/fs.h>
#include <linux/cdev.h>
#include <linux/device.h>
#include <linux/wait.h>
#include <linux/sched.h>
#include <linux/miscdevice.h>
#include <linux/workqueue.h>
#include <linux/uaccess.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/moduleparam.h>
#include <linux/err.h>
#include <linux/mtd/mtd.h>
#include <linux/sched.h>
#include <mach/hs_io_ctl_a.h>
#include <mach/smem_log.h>
#include "nand_FOTA.h"
#include "msm_nand.h"

#define MAX_MTD_DEVICES 32

#define DUMMY_PLTBL
#undef  DUMMY_PLTBL

#ifdef DUMMY_PLTBL
 #define ERR_ADDR 0xFFFFFFFF
 loff_t osbl_get_good_addr_num( loff_t physical_addr );
#else
 #include "hs_pl_table_proc.h"
#endif

#define PRINT_PREF KERN_INFO "nand_fota: "

static int dev;
static loff_t cnv_addr;
static int makPLTBLflg = false;

module_param(dev, int, S_IRUGO);
MODULE_PARM_DESC(dev, "MTD device number to use");

struct mtd_part {
    struct mtd_info mtd;
    struct mtd_info *master;
    uint64_t offset;
};

static struct mtd_info *mtd ;
static	struct mtd_part *part;

#define NAND_PAGE_SIZE 4096

#define PART(x)  ((struct mtd_part *)(x))

#define BOARD_NAND_BLOCK_SIZE (NAND_PAGE_SIZE*64)
#define FLASH_END_ADDR (1024*1024*1024-1)

struct part_info {
    uint64_t size;
    uint64_t offset;
};

static struct part_info part_tbl[ MAX_MTD_DEVICES ];

#define NAND_FOTA_DEBUG(...) printk(__VA_ARGS__)

struct fota_device_t {
    struct miscdevice misc;
};

static int fota_ioctl(struct inode *inode, struct file *file,
				unsigned int cmd, unsigned long arg)
{

    unsigned long result = -EFAULT;
    struct nand_fota_ioctl_info info;
    int offset;
    loff_t page_addr;
    unsigned char *mid_buff;
    loff_t physical_page_addr;

    switch(cmd)
    {
	case IOCTL_NAND_FOTA_INIT_CMD:
	    result = copy_from_user(&info,(struct nand_fota_ioctl_info __user *)arg,sizeof(info));
	    NAND_FOTA_DEBUG( "@@@DEBUG IOCTL_NAND_FOTA_INIT_CMD addr[%x] buf_addr[%x] size[%x]\n", (unsigned int)info.addr, (unsigned int)info.buff, info.size );

	    info.result = fota_flash_init();

	    NAND_FOTA_DEBUG("@@@DEBUG IOCTL_NAND_FOTA_INIT_CMD [%d]\n",info.result );
	    result = copy_to_user((struct nand_fota_ioctl_info __user *)arg,&info,sizeof(info));
	    break;

	case IOCTL_NAND_FOTA_READ_CMD:
	    mid_buff = (char*)kmalloc(4*64*1024, GFP_KERNEL);
	    if (!mid_buff)
	    {
		printk(KERN_ERR "fota_workbuf_read: Out of memory\n");
		info.size = 0;
	    }
	    else
	    {
		result = copy_from_user(&info,(struct nand_fota_ioctl_info __user *)arg,sizeof(info));
		NAND_FOTA_DEBUG("@@@DEBUG IOCTL_NAND_FOTA_READ_CMD addr[%#llx] buf_addr[%x] size[%x]\n", (long long)info.addr, (unsigned int)info.buff, info.size );
		if(((info.addr & (BOARD_NAND_BLOCK_SIZE-1)) + info.size ) > BOARD_NAND_BLOCK_SIZE)
		{
		    printk(KERN_ERR "Fota_Block_over_read\n");
		    info.size = 0;
		}
		else
        {
             physical_page_addr = PLTable_real_address( info.addr );
             if( physical_page_addr == HS_PL_ERR_ADDRESS )
             {
  		        printk(KERN_ERR "FOTA_PL_Tbl_error_read\n");
		        info.size = 0;
             }
             else
             {
		        offset = physical_page_addr & (NAND_PAGE_SIZE-1);
		        page_addr = physical_page_addr & ~(NAND_PAGE_SIZE-1);
		        NAND_FOTA_DEBUG("@@@DEBUG IOCTL_NAND_FOTA_READ_OFFSET addr[%#llx] offset[%x]\n", (long long)page_addr, offset );
		        info.result = fota_flash_read ( page_addr, mid_buff, info.size + offset );
		        if( info.result != 0 )
		        {
			        info.size = 0;
		        }
		        else
		        {
			        result = copy_to_user(info.buff, mid_buff+offset, info.size);
		        }

		    }
        }
		kfree(mid_buff);
	    }
	    NAND_FOTA_DEBUG("@@@DEBUG IOCTL_NAND_FOTA_READ_CMD [%d] [%d]\n",info.size, info.result);
	    result = copy_to_user((struct nand_fota_ioctl_info __user *)arg,&info,sizeof(info));
	    break;

	case IOCTL_NAND_FOTA_WRITE_CMD:
	    mid_buff = (char*)kmalloc(4*64*1024, GFP_KERNEL);
	    if (!mid_buff)
	    {
		printk(KERN_ERR "fota_workbuf_write: Out of memory\n");
		info.size = 0;
	    }
	    else
	    {
		result = copy_from_user(&info,(struct nand_fota_ioctl_info __user *)arg,sizeof(info));
		NAND_FOTA_DEBUG("@@@DEBUG IOCTL_NAND_FOTA_WRITE_CMD addr[%#llx] buf_addr[%x] size[%x]\n", (long long)info.addr, (unsigned int)info.buff, info.size);
		if(((info.addr & (BOARD_NAND_BLOCK_SIZE-1)) + info.size ) > BOARD_NAND_BLOCK_SIZE)
		{
		    printk(KERN_ERR "Fota_Block_over_write\n");
		    info.size = 0;
		}
        else
        {
            physical_page_addr = PLTable_real_address( info.addr );
            if( physical_page_addr == HS_PL_ERR_ADDRESS )
            {
  		       printk(KERN_ERR "FOTA_PL_Tbl_error_read\n");
		       info.size = 0;
            }
            else
            {
               while(physical_page_addr != HS_PL_ERR_RECOVERY)
               {
                   offset = physical_page_addr & (NAND_PAGE_SIZE-1);
                   page_addr = physical_page_addr & ~(NAND_PAGE_SIZE-1);
                   NAND_FOTA_DEBUG("@@@DEBUG IOCTL_NAND_FOTA_WRITE_OFFSET addr[%#llx] offset[%x]\n", (long long)page_addr, offset );
                   result = copy_from_user(mid_buff+offset, info.buff, info.size);
                   
                   info.result = fota_flash_write ( page_addr, mid_buff, info.size + offset );

                   if(info.result != 0 )
                   {
                       physical_page_addr = PLTable_bad_recovery( info.addr );
                       if(physical_page_addr == HS_PL_ERR_RECOVERY)
                       {
                           printk(KERN_ERR "PLTable_bad_recovery() faild\n");
                           info.size = 0;
                           break;
                       }
                   }
                   else
                   {
                       break;
                   }

                }
             }
          }
		
		  kfree(mid_buff);
	  }
	  NAND_FOTA_DEBUG("@@@DEBUG IOCTL_NAND_FOTA_WRITE_CMD [%x] [%d]\n",info.size, info.result );
	  result = copy_to_user((struct nand_fota_ioctl_info __user *)arg,&info,sizeof(info));
	  break;

	default:
    NAND_FOTA_DEBUG("@@@DEBUG default");
	    break;
    }
    return 0;
}

static int fota_open(struct inode *ip, struct file *fp)
{
    return nonseekable_open(ip, fp);
}

static int fota_release(struct inode *ip, struct file *fp)
{
    return 0;
}

static const struct file_operations fota_fops = {
    .owner = THIS_MODULE,
    .open = fota_open,
    .ioctl = fota_ioctl,
    .release = fota_release,
};

static struct fota_device_t fota_device = {
    .misc = {
	.minor = MISC_DYNAMIC_MINOR,
	.name = "fota_fl",
	.fops = &fota_fops,
    }
};

static void __exit nand_fota_exit(void)
{
    misc_deregister(&fota_device.misc);
}

static int __init nand_fota_init(void)
{
    int ret;
    ret = misc_register(&fota_device.misc);
    return ret;
}

module_init(nand_fota_init);
module_exit(nand_fota_exit);

MODULE_DESCRIPTION("NAND Memory FOTA Driver");
MODULE_LICENSE("GPL v2");


static int addr_check ( loff_t addr )
{

    int part_num;

    if( addr < part_tbl[ 0 ].offset || addr > FLASH_END_ADDR )
    {
	return -1;
    }	

    for( part_num=0; part_num<MAX_MTD_DEVICES; part_num++)
    {
	if( addr >= part_tbl[ part_num ].offset && addr < (part_tbl[ part_num ].offset + part_tbl[ part_num ].size))
	{
	    mtd = get_mtd_device(NULL, part_num);
	    cnv_addr = addr - part_tbl[ part_num ].offset;
	    return 0;
	}
	if( part_tbl[ part_num ].offset == 0 )
	{
	    return -1;
	}	
    }
    return -1;
}

static int  fota_flash_init ( void )
{

    printk(KERN_INFO "\n");
    printk(KERN_INFO "=================================================\n");
    printk(PRINT_PREF "MTD device: %d\n", dev);

    for(dev =0; dev<MAX_MTD_DEVICES; dev++)
    {
		part_tbl[ dev ].size = 0;	
		part_tbl[ dev ].offset = 0;	
    }

    for(dev =0; dev<MAX_MTD_DEVICES; dev++)
    {
		mtd = get_mtd_device(NULL, dev);

		if (IS_ERR(mtd)) {
		    printk(PRINT_PREF "error: cannot get MTD device\n");
		    break;;
		}

		if (mtd->type != MTD_NANDFLASH) {
		    printk(PRINT_PREF "this test requires NAND flash\n");
		    break;
		}
		part = PART(mtd);
		part_tbl[ dev ].size = mtd->size;	
		part_tbl[ dev ].offset = part->offset;	
    }
    printk(PRINT_PREF "fota_flash_init OK\n");
    printk(KERN_INFO "=================================================\n");
 
    if(makPLTBLflg == false){
        int ret = PLTable_table_setting();
        if(ret != 1){
            return -1;
        }
        makPLTBLflg = true;
    }
    return 0;

}


static int  fota_flash_read ( loff_t r_addr, unsigned char *r_buff, size_t r_size )
{


    int err=0;
    size_t  read_size = 0;
    int  cnv_size;

    if( addr_check( r_addr ) == -1 )
    {
        return -1;
    }
    cnv_size = r_size;

    while(cnv_size>0)
    {
        err = mtd->read( mtd, cnv_addr, NAND_PAGE_SIZE, &read_size, r_buff );

        if(err)
        {
             printk("%s: read_error %d\n",__func__, err);
             break;
        }
        printk("%s: read_size %x \n",__func__, read_size);
        cnv_size -= read_size;
        cnv_addr += read_size;
        r_buff   += read_size;
    }
    if (err == -EUCLEAN)
        err = 0;
    if (err) {
        printk(PRINT_PREF "error: read failed at %x %#llx %x\n",err ,(long long)r_addr ,read_size);
        return -1;
    }
    return 0;
}

static int  fota_flash_write ( loff_t w_addr, unsigned char *w_buff, size_t w_size )
{
    int err=0, status=0;
    size_t  write_size = 0;
    int  cnv_size;
    struct erase_info ei;

    if( addr_check( w_addr ) == -1 )
    {
        printk(PRINT_PREF "*******  wrirte_addr_check error:");
        return -1;
    }

    memset(&ei, 0, sizeof(struct erase_info));   
    ei.mtd  = mtd;
    ei.addr = cnv_addr & ~(BOARD_NAND_BLOCK_SIZE-1);
    ei.len  = mtd->erasesize;

    err = mtd->erase(mtd, &ei);

    if (err) {
	printk(PRINT_PREF "error %d while erasing \n", err);
	printk(PRINT_PREF "error %x while erasing2 \n", mtd->erasesize);
	status |= 0x0f;
    }

    if (ei.state == MTD_ERASE_FAILED) {
	printk(PRINT_PREF "some erase error occurred \n");
	status |= 0x0f0;
    }


    cnv_size = w_size;

    while(cnv_size>0)
    {
        err = mtd->write( mtd, cnv_addr, NAND_PAGE_SIZE, &write_size, w_buff );
        if(err)
        {
             break;
        }
        cnv_size -= write_size;
        cnv_addr += write_size;
        w_buff   += write_size;
    }

    if (err) {
        printk(PRINT_PREF "error: write failed at %x %#llx %x\n",err ,(long long)cnv_addr, write_size);
	status |= 0x0f00;
    }
    return status;
}

#ifdef DUMMY_PLTBL
loff_t osbl_get_good_addr_num( loff_t physical_addr )
{
   return physical_addr & ~( BOARD_NAND_BLOCK_SIZE-1 );
}
#endif

